#
# -*- coding: utf-8 -*-
#
# This file is generated using
#   ./script/generate-geo-data > octodns/record/geo_data.py
# do not modify it directly
#

geo_data = \
    {'AF': {'AO': {'name': 'Angola'},
            'BF': {'name': 'Burkina Faso'},
            'BI': {'name': 'Burundi'},
            'BJ': {'name': 'Benin'},
            'BW': {'name': 'Botswana'},
            'CD': {'name': 'Congo, The Democratic Republic of the'},
            'CF': {'name': 'Central African Republic'},
            'CG': {'name': 'Congo'},
            'CI': {'name': "Côte d'Ivoire"},
            'CM': {'name': 'Cameroon'},
            'CV': {'name': 'Cabo Verde'},
            'DJ': {'name': 'Djibouti'},
            'DZ': {'name': 'Algeria'},
            'EG': {'name': 'Egypt'},
            'EH': {'name': 'Western Sahara'},
            'ER': {'name': 'Eritrea'},
            'ET': {'name': 'Ethiopia'},
            'GA': {'name': 'Gabon'},
            'GH': {'name': 'Ghana'},
            'GM': {'name': 'Gambia'},
            'GN': {'name': 'Guinea'},
            'GQ': {'name': 'Equatorial Guinea'},
            'GW': {'name': 'Guinea-Bissau'},
            'KE': {'name': 'Kenya'},
            'KM': {'name': 'Comoros'},
            'LR': {'name': 'Liberia'},
            'LS': {'name': 'Lesotho'},
            'LY': {'name': 'Libya'},
            'MA': {'name': 'Morocco'},
            'MG': {'name': 'Madagascar'},
            'ML': {'name': 'Mali'},
            'MR': {'name': 'Mauritania'},
            'MU': {'name': 'Mauritius'},
            'MW': {'name': 'Malawi'},
            'MZ': {'name': 'Mozambique'},
            'NA': {'name': 'Namibia'},
            'NE': {'name': 'Niger'},
            'NG': {'name': 'Nigeria'},
            'RE': {'name': 'Réunion'},
            'RW': {'name': 'Rwanda'},
            'SC': {'name': 'Seychelles'},
            'SD': {'name': 'Sudan'},
            'SH': {'name': 'Saint Helena, Ascension and Tristan da Cunha'},
            'SL': {'name': 'Sierra Leone'},
            'SN': {'name': 'Senegal'},
            'SO': {'name': 'Somalia'},
            'SS': {'name': 'South Sudan'},
            'ST': {'name': 'Sao Tome and Principe'},
            'SZ': {'name': 'Swaziland'},
            'TD': {'name': 'Chad'},
            'TG': {'name': 'Togo'},
            'TN': {'name': 'Tunisia'},
            'TZ': {'name': 'Tanzania, United Republic of'},
            'UG': {'name': 'Uganda'},
            'YT': {'name': 'Mayotte'},
            'ZA': {'name': 'South Africa'},
            'ZM': {'name': 'Zambia'},
            'ZW': {'name': 'Zimbabwe'}},
     'AN': {'AQ': {'name': 'Antarctica'},
            'BV': {'name': 'Bouvet Island'},
            'HM': {'name': 'Heard Island and McDonald Islands'},
            'TF': {'name': 'French Southern Territories'}},
     'AS': {'AE': {'name': 'United Arab Emirates'},
            'AF': {'name': 'Afghanistan'},
            'AM': {'name': 'Armenia'},
            'AZ': {'name': 'Azerbaijan'},
            'BD': {'name': 'Bangladesh'},
            'BH': {'name': 'Bahrain'},
            'BN': {'name': 'Brunei Darussalam'},
            'BT': {'name': 'Bhutan'},
            'CC': {'name': 'Cocos (Keeling) Islands'},
            'CN': {'name': 'China'},
            'CX': {'name': 'Christmas Island'},
            'CY': {'name': 'Cyprus'},
            'GE': {'name': 'Georgia'},
            'HK': {'name': 'Hong Kong'},
            'ID': {'name': 'Indonesia'},
            'IL': {'name': 'Israel'},
            'IN': {'name': 'India'},
            'IO': {'name': 'British Indian Ocean Territory'},
            'IQ': {'name': 'Iraq'},
            'IR': {'name': 'Iran, Islamic Republic of'},
            'JO': {'name': 'Jordan'},
            'JP': {'name': 'Japan'},
            'KG': {'name': 'Kyrgyzstan'},
            'KH': {'name': 'Cambodia'},
            'KP': {'name': "Korea, Democratic People's Republic of"},
            'KR': {'name': 'Korea, Republic of'},
            'KW': {'name': 'Kuwait'},
            'KZ': {'name': 'Kazakhstan'},
            'LA': {'name': "Lao People's Democratic Republic"},
            'LB': {'name': 'Lebanon'},
            'LK': {'name': 'Sri Lanka'},
            'MM': {'name': 'Myanmar'},
            'MN': {'name': 'Mongolia'},
            'MO': {'name': 'Macao'},
            'MV': {'name': 'Maldives'},
            'MY': {'name': 'Malaysia'},
            'NP': {'name': 'Nepal'},
            'OM': {'name': 'Oman'},
            'PH': {'name': 'Philippines'},
            'PK': {'name': 'Pakistan'},
            'PS': {'name': 'Palestine, State of'},
            'QA': {'name': 'Qatar'},
            'SA': {'name': 'Saudi Arabia'},
            'SG': {'name': 'Singapore'},
            'SY': {'name': 'Syrian Arab Republic'},
            'TH': {'name': 'Thailand'},
            'TJ': {'name': 'Tajikistan'},
            'TM': {'name': 'Turkmenistan'},
            'TR': {'name': 'Turkey'},
            'TW': {'name': 'Taiwan, Province of China'},
            'UZ': {'name': 'Uzbekistan'},
            'VN': {'name': 'Viet Nam'},
            'YE': {'name': 'Yemen'}},
     'EU': {'AD': {'name': 'Andorra'},
            'AL': {'name': 'Albania'},
            'AT': {'name': 'Austria'},
            'AX': {'name': 'Åland Islands'},
            'BA': {'name': 'Bosnia and Herzegovina'},
            'BE': {'name': 'Belgium'},
            'BG': {'name': 'Bulgaria'},
            'BY': {'name': 'Belarus'},
            'CH': {'name': 'Switzerland'},
            'CZ': {'name': 'Czechia'},
            'DE': {'name': 'Germany'},
            'DK': {'name': 'Denmark'},
            'EE': {'name': 'Estonia'},
            'ES': {'name': 'Spain'},
            'FI': {'name': 'Finland'},
            'FO': {'name': 'Faroe Islands'},
            'FR': {'name': 'France'},
            'GB': {'name': 'United Kingdom'},
            'GG': {'name': 'Guernsey'},
            'GI': {'name': 'Gibraltar'},
            'GR': {'name': 'Greece'},
            'HR': {'name': 'Croatia'},
            'HU': {'name': 'Hungary'},
            'IE': {'name': 'Ireland'},
            'IM': {'name': 'Isle of Man'},
            'IS': {'name': 'Iceland'},
            'IT': {'name': 'Italy'},
            'JE': {'name': 'Jersey'},
            'LI': {'name': 'Liechtenstein'},
            'LT': {'name': 'Lithuania'},
            'LU': {'name': 'Luxembourg'},
            'LV': {'name': 'Latvia'},
            'MC': {'name': 'Monaco'},
            'MD': {'name': 'Moldova, Republic of'},
            'ME': {'name': 'Montenegro'},
            'MK': {'name': 'Macedonia, Republic of'},
            'MT': {'name': 'Malta'},
            'NL': {'name': 'Netherlands'},
            'NO': {'name': 'Norway'},
            'PL': {'name': 'Poland'},
            'PT': {'name': 'Portugal'},
            'RO': {'name': 'Romania'},
            'RS': {'name': 'Serbia'},
            'RU': {'name': 'Russian Federation'},
            'SE': {'name': 'Sweden'},
            'SI': {'name': 'Slovenia'},
            'SJ': {'name': 'Svalbard and Jan Mayen'},
            'SK': {'name': 'Slovakia'},
            'SM': {'name': 'San Marino'},
            'UA': {'name': 'Ukraine'},
            'VA': {'name': 'Holy See (Vatican City State)'}},
     'ID': {'TL': {'name': 'Timor-Leste'}},
     'NA': {'AG': {'name': 'Antigua and Barbuda'},
            'AI': {'name': 'Anguilla'},
            'AW': {'name': 'Aruba'},
            'BB': {'name': 'Barbados'},
            'BL': {'name': 'Saint Barthélemy'},
            'BM': {'name': 'Bermuda'},
            'BQ': {'name': 'Bonaire, Sint Eustatius and Saba'},
            'BS': {'name': 'Bahamas'},
            'BZ': {'name': 'Belize'},
            'CA': {'name': 'Canada'},
            'CR': {'name': 'Costa Rica'},
            'CU': {'name': 'Cuba'},
            'CW': {'name': 'Curaçao'},
            'DM': {'name': 'Dominica'},
            'DO': {'name': 'Dominican Republic'},
            'GD': {'name': 'Grenada'},
            'GL': {'name': 'Greenland'},
            'GP': {'name': 'Guadeloupe'},
            'GT': {'name': 'Guatemala'},
            'HN': {'name': 'Honduras'},
            'HT': {'name': 'Haiti'},
            'JM': {'name': 'Jamaica'},
            'KN': {'name': 'Saint Kitts and Nevis'},
            'KY': {'name': 'Cayman Islands'},
            'LC': {'name': 'Saint Lucia'},
            'MF': {'name': 'Saint Martin (French part)'},
            'MQ': {'name': 'Martinique'},
            'MS': {'name': 'Montserrat'},
            'MX': {'name': 'Mexico'},
            'NI': {'name': 'Nicaragua'},
            'PA': {'name': 'Panama'},
            'PM': {'name': 'Saint Pierre and Miquelon'},
            'PR': {'name': 'Puerto Rico'},
            'SV': {'name': 'El Salvador'},
            'SX': {'name': 'Sint Maarten (Dutch part)'},
            'TC': {'name': 'Turks and Caicos Islands'},
            'TT': {'name': 'Trinidad and Tobago'},
            'US': {'name': 'United States',
                   'provinces': {'AK': {'name': 'Alaska'},
                                 'AL': {'name': 'Alabama'},
                                 'AR': {'name': 'Arkansas'},
                                 'AS': {'name': 'American Samoa'},
                                 'AZ': {'name': 'Arizona'},
                                 'CA': {'name': 'California'},
                                 'CO': {'name': 'Colorado'},
                                 'CT': {'name': 'Connecticut'},
                                 'DC': {'name': 'District of Columbia'},
                                 'DE': {'name': 'Delaware'},
                                 'FL': {'name': 'Florida'},
                                 'GA': {'name': 'Georgia'},
                                 'GU': {'name': 'Guam'},
                                 'HI': {'name': 'Hawaii'},
                                 'IA': {'name': 'Iowa'},
                                 'ID': {'name': 'Idaho'},
                                 'IL': {'name': 'Illinois'},
                                 'IN': {'name': 'Indiana'},
                                 'KS': {'name': 'Kansas'},
                                 'KY': {'name': 'Kentucky'},
                                 'LA': {'name': 'Louisiana'},
                                 'MA': {'name': 'Massachusetts'},
                                 'MD': {'name': 'Maryland'},
                                 'ME': {'name': 'Maine'},
                                 'MI': {'name': 'Michigan'},
                                 'MN': {'name': 'Minnesota'},
                                 'MO': {'name': 'Missouri'},
                                 'MP': {'name': 'Northern Mariana Islands'},
                                 'MS': {'name': 'Mississippi'},
                                 'MT': {'name': 'Montana'},
                                 'NC': {'name': 'North Carolina'},
                                 'ND': {'name': 'North Dakota'},
                                 'NE': {'name': 'Nebraska'},
                                 'NH': {'name': 'New Hampshire'},
                                 'NJ': {'name': 'New Jersey'},
                                 'NM': {'name': 'New Mexico'},
                                 'NV': {'name': 'Nevada'},
                                 'NY': {'name': 'New York'},
                                 'OH': {'name': 'Ohio'},
                                 'OK': {'name': 'Oklahoma'},
                                 'OR': {'name': 'Oregon'},
                                 'PA': {'name': 'Pennsylvania'},
                                 'PR': {'name': 'Puerto Rico'},
                                 'RI': {'name': 'Rhode Island'},
                                 'SC': {'name': 'South Carolina'},
                                 'SD': {'name': 'South Dakota'},
                                 'TN': {'name': 'Tennessee'},
                                 'TX': {'name': 'Texas'},
                                 'UM': {'name': 'United States Minor Outlying '
                                                'Islands'},
                                 'UT': {'name': 'Utah'},
                                 'VA': {'name': 'Virginia'},
                                 'VI': {'name': 'Virgin Islands'},
                                 'VT': {'name': 'Vermont'},
                                 'WA': {'name': 'Washington'},
                                 'WI': {'name': 'Wisconsin'},
                                 'WV': {'name': 'West Virginia'},
                                 'WY': {'name': 'Wyoming'}}},
            'VC': {'name': 'Saint Vincent and the Grenadines'},
            'VG': {'name': 'Virgin Islands, British'},
            'VI': {'name': 'Virgin Islands, U.S.'}},
     'OC': {'AS': {'name': 'American Samoa'},
            'AU': {'name': 'Australia'},
            'CK': {'name': 'Cook Islands'},
            'FJ': {'name': 'Fiji'},
            'FM': {'name': 'Micronesia, Federated States of'},
            'GU': {'name': 'Guam'},
            'KI': {'name': 'Kiribati'},
            'MH': {'name': 'Marshall Islands'},
            'MP': {'name': 'Northern Mariana Islands'},
            'NC': {'name': 'New Caledonia'},
            'NF': {'name': 'Norfolk Island'},
            'NR': {'name': 'Nauru'},
            'NU': {'name': 'Niue'},
            'NZ': {'name': 'New Zealand'},
            'PF': {'name': 'French Polynesia'},
            'PG': {'name': 'Papua New Guinea'},
            'PN': {'name': 'Pitcairn'},
            'PW': {'name': 'Palau'},
            'SB': {'name': 'Solomon Islands'},
            'TK': {'name': 'Tokelau'},
            'TO': {'name': 'Tonga'},
            'TV': {'name': 'Tuvalu'},
            'UM': {'name': 'United States Minor Outlying Islands'},
            'VU': {'name': 'Vanuatu'},
            'WF': {'name': 'Wallis and Futuna'},
            'WS': {'name': 'Samoa'}},
     'SA': {'AR': {'name': 'Argentina'},
            'BO': {'name': 'Bolivia, Plurinational State of'},
            'BR': {'name': 'Brazil'},
            'CL': {'name': 'Chile'},
            'CO': {'name': 'Colombia'},
            'EC': {'name': 'Ecuador'},
            'FK': {'name': 'Falkland Islands (Malvinas)'},
            'GF': {'name': 'French Guiana'},
            'GS': {'name': 'South Georgia and the South Sandwich Islands'},
            'GY': {'name': 'Guyana'},
            'PE': {'name': 'Peru'},
            'PY': {'name': 'Paraguay'},
            'SR': {'name': 'Suriname'},
            'UY': {'name': 'Uruguay'},
            'VE': {'name': 'Venezuela, Bolivarian Republic of'}}}



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

import dns.zone
from dns.exception import DNSException

from mock import patch
from unittest import TestCase

from octodns.source.axfr import AxfrSource, AxfrSourceZoneTransferFailed, \
    ZoneFileSource, ZoneFileSourceLoadFailure
from octodns.zone import Zone


class TestAxfrSource(TestCase):
    source = AxfrSource('test', 'localhost')

    forward_zonefile = dns.zone.from_file('./tests/zones/unit.tests.',
                                          'unit.tests', relativize=False)

    @patch('dns.zone.from_xfr')
    def test_populate(self, from_xfr_mock):
        got = Zone('unit.tests.', [])

        from_xfr_mock.side_effect = [
            self.forward_zonefile,
            DNSException
        ]

        self.source.populate(got)
        self.assertEquals(11, len(got.records))

        with self.assertRaises(AxfrSourceZoneTransferFailed) as ctx:
            zone = Zone('unit.tests.', [])
            self.source.populate(zone)
        self.assertEquals('Unable to Perform Zone Transfer',
                          ctx.exception.message)


class TestZoneFileSource(TestCase):
    source = ZoneFileSource('test', './tests/zones')

    def test_populate(self):
        # Valid zone file in directory
        valid = Zone('unit.tests.', [])
        self.source.populate(valid)
        self.assertEquals(11, len(valid.records))

        # 2nd populate does not read file again
        again = Zone('unit.tests.', [])
        self.source.populate(again)
        self.assertEquals(11, len(again.records))

        # bust the cache
        del self.source._zone_records[valid.name]

        # No zone file in directory
        missing = Zone('missing.zone.', [])
        self.source.populate(missing)
        self.assertEquals(0, len(missing.records))

        # Zone file is not valid
        with self.assertRaises(ZoneFileSourceLoadFailure) as ctx:
            zone = Zone('invalid.zone.', [])
            self.source.populate(zone)
        self.assertEquals('The DNS zone has no NS RRset at its origin.',
                          ctx.exception.message)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from octodns.record import Create, Delete, Update, Record
from octodns.provider.googlecloud import GoogleCloudProvider

from octodns.zone import Zone
from octodns.provider.base import Plan, BaseProvider

from unittest import TestCase
from mock import Mock, patch, PropertyMock

zone = Zone(name='unit.tests.', sub_zones=[])
octo_records = []
octo_records.append(Record.new(zone, '', {
    'ttl': 0,
    'type': 'A',
    'values': ['1.2.3.4', '10.10.10.10']}))
octo_records.append(Record.new(zone, 'a', {
    'ttl': 1,
    'type': 'A',
    'values': ['1.2.3.4', '1.1.1.1']}))
octo_records.append(Record.new(zone, 'aa', {
    'ttl': 9001,
    'type': 'A',
    'values': ['1.2.4.3']}))
octo_records.append(Record.new(zone, 'aaa', {
    'ttl': 2,
    'type': 'A',
    'values': ['1.1.1.3']}))
octo_records.append(Record.new(zone, 'cname', {
    'ttl': 3,
    'type': 'CNAME',
    'value': 'a.unit.tests.'}))
octo_records.append(Record.new(zone, 'mx1', {
    'ttl': 3,
    'type': 'MX',
    'values': [{
        'priority': 10,
        'value': 'mx1.unit.tests.',
    }, {
        'priority': 20,
        'value': 'mx2.unit.tests.',
    }]}))
octo_records.append(Record.new(zone, 'mx2', {
    'ttl': 3,
    'type': 'MX',
    'values': [{
        'priority': 10,
        'value': 'mx1.unit.tests.',
    }]}))
octo_records.append(Record.new(zone, '', {
    'ttl': 4,
    'type': 'NS',
    'values': ['ns1.unit.tests.', 'ns2.unit.tests.']}))
octo_records.append(Record.new(zone, 'foo', {
    'ttl': 5,
    'type': 'NS',
    'value': 'ns1.unit.tests.'}))
octo_records.append(Record.new(zone, '_srv._tcp', {
    'ttl': 6,
    'type': 'SRV',
    'values': [{
        'priority': 10,
        'weight': 20,
        'port': 30,
        'target': 'foo-1.unit.tests.',
    }, {
        'priority': 12,
        'weight': 30,
        'port': 30,
        'target': 'foo-2.unit.tests.',
    }]}))
octo_records.append(Record.new(zone, '_srv2._tcp', {
    'ttl': 7,
    'type': 'SRV',
    'values': [{
        'priority': 12,
        'weight': 17,
        'port': 1,
        'target': 'srvfoo.unit.tests.',
    }]}))
octo_records.append(Record.new(zone, 'txt1', {
    'ttl': 8,
    'type': 'TXT',
    'value': 'txt singleton test'}))
octo_records.append(Record.new(zone, 'txt2', {
    'ttl': 9,
    'type': 'TXT',
    'values': ['txt multiple test', 'txt multiple test 2']}))
octo_records.append(Record.new(zone, 'naptr', {
    'ttl': 9,
    'type': 'NAPTR',
    'values': [{
        'order': 100,
        'preference': 10,
        'flags': 'S',
        'service': 'SIP+D2U',
        'regexp': "!^.*$!sip:customer-service@unit.tests!",
        'replacement': '_sip._udp.unit.tests.'
    }]}))
octo_records.append(Record.new(zone, 'caa', {
    'ttl': 9,
    'type': 'CAA',
    'value': {
        'flags': 0,
        'tag': 'issue',
        'value': 'ca.unit.tests',
    }}))
for record in octo_records:
    zone.add_record(record)

# This is the format which the google API likes.
resource_record_sets = [
    ('unit.tests.', u'A', 0, [u'1.2.3.4', u'10.10.10.10']),
    (u'a.unit.tests.', u'A', 1, [u'1.1.1.1', u'1.2.3.4']),
    (u'aa.unit.tests.', u'A', 9001, [u'1.2.4.3']),
    (u'aaa.unit.tests.', u'A', 2, [u'1.1.1.3']),
    (u'cname.unit.tests.', u'CNAME', 3, [u'a.unit.tests.']),
    (u'mx1.unit.tests.', u'MX', 3,
     [u'10 mx1.unit.tests.', u'20 mx2.unit.tests.']),
    (u'mx2.unit.tests.', u'MX', 3, [u'10 mx1.unit.tests.']),
    ('unit.tests.', u'NS', 4, [u'ns1.unit.tests.', u'ns2.unit.tests.']),
    (u'foo.unit.tests.', u'NS', 5, [u'ns1.unit.tests.']),
    (u'_srv._tcp.unit.tests.', u'SRV', 6,
     [u'10 20 30 foo-1.unit.tests.', u'12 30 30 foo-2.unit.tests.']),
    (u'_srv2._tcp.unit.tests.', u'SRV', 7, [u'12 17 1 srvfoo.unit.tests.']),
    (u'txt1.unit.tests.', u'TXT', 8, [u'txt singleton test']),
    (u'txt2.unit.tests.', u'TXT', 9,
     [u'txt multiple test', u'txt multiple test 2']),
    (u'naptr.unit.tests.', u'NAPTR', 9, [
        u'100 10 "S" "SIP+D2U" "!^.*$!sip:customer-service@unit.tests!"'
        u' _sip._udp.unit.tests.']),
    (u'caa.unit.tests.', u'CAA', 9, [u'0 issue ca.unit.tests'])
]


class DummyResourceRecordSet:
    def __init__(self, record_name, record_type, ttl, rrdatas):
        self.name = record_name
        self.record_type = record_type
        self.ttl = ttl
        self.rrdatas = rrdatas

    def __eq__(self, other):
        try:
            return self.name == other.name \
                and self.record_type == other.record_type \
                and self.ttl == other.ttl  \
                and sorted(self.rrdatas) == sorted(other.rrdatas)
        except:
            return False

    def __repr__(self):
        return "{} {} {} {!s}"\
            .format(self.name, self.record_type, self.ttl, self.rrdatas)

    def __hash__(self):
        return hash(repr(self))


class DummyGoogleCloudZone:
    def __init__(self, dns_name, name=""):
        self.dns_name = dns_name
        self.name = name

    def resource_record_set(self, *args):
        return DummyResourceRecordSet(*args)

    def list_resource_record_sets(self, *args):
        pass

    def create(self, *args, **kwargs):
        pass


class DummyIterator:
    """Returns a mock DummyIterator object to use in testing.
    This is because API calls for google cloud DNS, if paged, contains a
    "next_page_token", which can be used to grab a subsequent
    iterator with more results.

        :type return: DummyIterator
    """
    def __init__(self,  list_of_stuff, page_token=None):
        self.iterable = iter(list_of_stuff)
        self.next_page_token = page_token

    def __iter__(self):
        return self

    def next(self):
        return self.iterable.next()


class TestGoogleCloudProvider(TestCase):
    @patch('octodns.provider.googlecloud.dns')
    def _get_provider(*args):
        '''Returns a mock GoogleCloudProvider object to use in testing.

            :type return: GoogleCloudProvider
        '''
        return GoogleCloudProvider(id=1, project="mock")

    @patch('octodns.provider.googlecloud.dns')
    def test___init__(self, *_):
        self.assertIsInstance(GoogleCloudProvider(id=1,
                                                  credentials_file="test",
                                                  project="unit test"),
                              BaseProvider)

        self.assertIsInstance(GoogleCloudProvider(id=1),
                              BaseProvider)

    @patch('octodns.provider.googlecloud.time.sleep')
    @patch('octodns.provider.googlecloud.dns')
    def test__apply(self, *_):
        class DummyDesired:
            def __init__(self, name, changes):
                self.name = name
                self.changes = changes

        apply_z = Zone("unit.tests.", [])
        create_r = Record.new(apply_z, '', {
            'ttl': 0,
            'type': 'A',
            'values': ['1.2.3.4', '10.10.10.10']})
        delete_r = Record.new(apply_z, 'a', {
            'ttl': 1,
            'type': 'A',
            'values': ['1.2.3.4', '1.1.1.1']})
        update_existing_r = Record.new(apply_z, 'aa', {
            'ttl': 9001,
            'type': 'A',
            'values': ['1.2.4.3']})
        update_new_r = Record.new(apply_z, 'aa', {
            'ttl': 666,
            'type': 'A',
            'values': ['1.4.3.2']})

        gcloud_zone_mock = DummyGoogleCloudZone("unit.tests.", "unit-tests")
        status_mock = Mock()
        return_values_for_status = iter(
            ["pending"] * 11 + ['done', 'done'])
        type(status_mock).status = PropertyMock(
            side_effect=return_values_for_status.next)
        gcloud_zone_mock.changes = Mock(return_value=status_mock)

        provider = self._get_provider()
        provider.gcloud_client = Mock()
        provider._gcloud_zones = {"unit.tests.": gcloud_zone_mock}
        desired = Mock()
        desired.name = "unit.tests."
        changes = []
        changes.append(Create(create_r))
        changes.append(Delete(delete_r))
        changes.append(Update(existing=update_existing_r, new=update_new_r))

        provider.apply(Plan(
            existing=[update_existing_r, delete_r],
            desired=desired,
            changes=changes,
            exists=True
        ))

        calls_mock = gcloud_zone_mock.changes.return_value
        mocked_calls = []
        for mock_call in calls_mock.add_record_set.mock_calls:
            mocked_calls.append(mock_call[1][0])

        self.assertEqual(mocked_calls, [
            DummyResourceRecordSet(
                'unit.tests.', 'A', 0, ['1.2.3.4', '10.10.10.10']),
            DummyResourceRecordSet(
                'aa.unit.tests.', 'A', 666, ['1.4.3.2'])
        ])

        mocked_calls2 = []
        for mock_call in calls_mock.delete_record_set.mock_calls:
            mocked_calls2.append(mock_call[1][0])

        self.assertEqual(mocked_calls2, [
            DummyResourceRecordSet(
                'a.unit.tests.', 'A', 1, ['1.2.3.4', '1.1.1.1']),
            DummyResourceRecordSet(
                'aa.unit.tests.', 'A', 9001, ['1.2.4.3'])
        ])

        type(status_mock).status = "pending"

        with self.assertRaises(RuntimeError):
            provider.apply(Plan(
                existing=[update_existing_r, delete_r],
                desired=desired,
                changes=changes,
                exists=True
            ))

        unsupported_change = Mock()
        unsupported_change.__len__ = Mock(return_value=1)
        type_mock = Mock()
        type_mock._type = "A"
        unsupported_change.record = type_mock

        mock_plan = Mock()
        type(mock_plan).desired = PropertyMock(return_value=DummyDesired(
            "dummy name", []))
        type(mock_plan).changes = [unsupported_change]

        with self.assertRaises(RuntimeError):
            provider.apply(mock_plan)

    def test__get_gcloud_client(self):
        provider = self._get_provider()

        self.assertIsInstance(provider, GoogleCloudProvider)

    @patch('octodns.provider.googlecloud.dns')
    def test_populate(self, _):
        def _get_mock_zones(page_token=None):
            if not page_token:
                return DummyIterator([
                    DummyGoogleCloudZone('example.com.'),
                ], page_token="MOCK_PAGE_TOKEN")
            elif page_token == "MOCK_PAGE_TOKEN":
                return DummyIterator([
                    DummyGoogleCloudZone('example2.com.'),
                ], page_token="MOCK_PAGE_TOKEN2")

            return DummyIterator([
                google_cloud_zone
            ])

        def _get_mock_record_sets(page_token=None):
            if not page_token:
                return DummyIterator(
                    [DummyResourceRecordSet(*v) for v in
                     resource_record_sets[:3]], page_token="MOCK_PAGE_TOKEN")
            elif page_token == "MOCK_PAGE_TOKEN":

                return DummyIterator(
                    [DummyResourceRecordSet(*v) for v in
                     resource_record_sets[3:5]], page_token="MOCK_PAGE_TOKEN2")
            return DummyIterator(
                [DummyResourceRecordSet(*v) for v in resource_record_sets[5:]])

        google_cloud_zone = DummyGoogleCloudZone('unit.tests.')

        provider = self._get_provider()
        provider.gcloud_client.list_zones = Mock(side_effect=_get_mock_zones)
        google_cloud_zone.list_resource_record_sets = Mock(
            side_effect=_get_mock_record_sets)

        self.assertEqual(provider.gcloud_zones.get("unit.tests.").dns_name,
                         "unit.tests.")

        test_zone = Zone('unit.tests.', [])
        exists = provider.populate(test_zone)
        self.assertTrue(exists)

        # test_zone gets fed the same records as zone does, except it's in
        # the format returned by google API, so after populate they should look
        # exactly the same.
        self.assertEqual(test_zone.records, zone.records)

        test_zone2 = Zone('nonexistent.zone.', [])
        exists = provider.populate(test_zone2, False, False)
        self.assertFalse(exists)

        self.assertEqual(len(test_zone2.records), 0,
                         msg="Zone should not get records from wrong domain")

        provider.SUPPORTS = set()
        test_zone3 = Zone('unit.tests.', [])
        provider.populate(test_zone3)
        self.assertEqual(len(test_zone3.records), 0)

    @patch('octodns.provider.googlecloud.dns')
    def test_populate_corner_cases(self, _):
        provider = self._get_provider()
        test_zone = Zone('unit.tests.', [])
        not_same_fqdn = DummyResourceRecordSet(
            'unit.tests.gr', u'A', 0, [u'1.2.3.4']),

        provider._get_gcloud_records = Mock(
            side_effect=[not_same_fqdn])
        provider._gcloud_zones = {
            "unit.tests.": DummyGoogleCloudZone("unit.tests.", "unit-tests")}

        provider.populate(test_zone)

        self.assertEqual(len(test_zone.records), 1)

        self.assertEqual(test_zone.records.pop().fqdn,
                         u'unit.tests.gr.unit.tests.')

    def test__get_gcloud_zone(self):
        provider = self._get_provider()

        provider.gcloud_client = Mock()
        provider.gcloud_client.list_zones = Mock(
            return_value=DummyIterator([]))

        self.assertIsNone(provider.gcloud_zones.get("nonexistent.zone"),
                          msg="Check that nonexistent zones return None when"
                              "there's no create=True flag")

    def test__get_rrsets(self):
        provider = self._get_provider()
        dummy_gcloud_zone = DummyGoogleCloudZone("unit.tests")
        for octo_record in octo_records:
            _rrset_func = getattr(
                provider, '_rrset_for_{}'.format(octo_record._type))
            self.assertEqual(
                _rrset_func(dummy_gcloud_zone, octo_record).record_type,
                octo_record._type
            )

    def test__create_zone(self):
        provider = self._get_provider()

        provider.gcloud_client = Mock()
        provider.gcloud_client.list_zones = Mock(
            return_value=DummyIterator([]))

        mock_zone = provider._create_gcloud_zone("nonexistent.zone.mock")

        mock_zone.create.assert_called()
        provider.gcloud_client.zone.assert_called()

    def test__create_zone_ip6_arpa(self):
        def _create_dummy_zone(name, dns_name):
            return DummyGoogleCloudZone(name=name, dns_name=dns_name)

        provider = self._get_provider()

        provider.gcloud_client = Mock()
        provider.gcloud_client.zone = Mock(side_effect=_create_dummy_zone)

        mock_zone = \
            provider._create_gcloud_zone('0.0.0.0.8.b.d.0.1.0.0.2.ip6.arpa')

        self.assertRegexpMatches(mock_zone.name, '^[a-z][a-z0-9-]*[a-z0-9]$')
        self.assertEqual(len(mock_zone.name), 63)

    def test_semicolon_fixup(self):
        provider = self._get_provider()

        self.assertEquals({
            'values': ['abcd\\; ef\\;g', 'hij\\; klm\\;n']
        }, provider._data_for_TXT(
            DummyResourceRecordSet(
                'unit.tests.', 'TXT', 0, ['abcd; ef;g', 'hij\\; klm\\;n'])
        ))



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from os import makedirs
from os.path import basename, dirname, isdir, isfile, join
from unittest import TestCase
from yaml import safe_load
from yaml.constructor import ConstructorError

from octodns.record import Create
from octodns.provider.base import Plan
from octodns.provider.yaml import _list_all_yaml_files, \
    SplitYamlProvider, YamlProvider
from octodns.zone import SubzoneRecordException, Zone

from helpers import TemporaryDirectory


class TestYamlProvider(TestCase):

    def test_provider(self):
        source = YamlProvider('test', join(dirname(__file__), 'config'))

        zone = Zone('unit.tests.', [])
        dynamic_zone = Zone('dynamic.tests.', [])

        # With target we don't add anything
        source.populate(zone, target=source)
        self.assertEquals(0, len(zone.records))

        # without it we see everything
        source.populate(zone)
        self.assertEquals(18, len(zone.records))

        source.populate(dynamic_zone)
        self.assertEquals(5, len(dynamic_zone.records))

        # Assumption here is that a clean round-trip means that everything
        # worked as expected, data that went in came back out and could be
        # pulled in yet again and still match up. That assumes that the input
        # data completely exercises things. This assumption can be tested by
        # relatively well by running
        #   ./script/coverage tests/test_octodns_provider_yaml.py and
        # looking at the coverage file
        #   ./htmlcov/octodns_provider_yaml_py.html

        with TemporaryDirectory() as td:
            # Add some subdirs to make sure that it can create them
            directory = join(td.dirname, 'sub', 'dir')
            yaml_file = join(directory, 'unit.tests.yaml')
            dynamic_yaml_file = join(directory, 'dynamic.tests.yaml')
            target = YamlProvider('test', directory)

            # We add everything
            plan = target.plan(zone)
            self.assertEquals(15, len(filter(lambda c: isinstance(c, Create),
                                             plan.changes)))
            self.assertFalse(isfile(yaml_file))

            # Now actually do it
            self.assertEquals(15, target.apply(plan))
            self.assertTrue(isfile(yaml_file))

            # Dynamic plan
            plan = target.plan(dynamic_zone)
            self.assertEquals(5, len(filter(lambda c: isinstance(c, Create),
                                            plan.changes)))
            self.assertFalse(isfile(dynamic_yaml_file))
            # Apply it
            self.assertEquals(5, target.apply(plan))
            self.assertTrue(isfile(dynamic_yaml_file))

            # There should be no changes after the round trip
            reloaded = Zone('unit.tests.', [])
            target.populate(reloaded)
            self.assertDictEqual(
                {'included': ['test']},
                filter(
                    lambda x: x.name == 'included', reloaded.records
                )[0]._octodns)

            self.assertFalse(zone.changes(reloaded, target=source))

            # A 2nd sync should still create everything
            plan = target.plan(zone)
            self.assertEquals(15, len(filter(lambda c: isinstance(c, Create),
                                             plan.changes)))

            with open(yaml_file) as fh:
                data = safe_load(fh.read())

                # '' has some of both
                roots = sorted(data.pop(''), key=lambda r: r['type'])
                self.assertTrue('values' in roots[0])  # A
                self.assertTrue('geo' in roots[0])  # geo made the trip
                self.assertTrue('value' in roots[1])   # CAA
                self.assertTrue('values' in roots[2])  # SSHFP

                # these are stored as plural 'values'
                self.assertTrue('values' in data.pop('_srv._tcp'))
                self.assertTrue('values' in data.pop('mx'))
                self.assertTrue('values' in data.pop('naptr'))
                self.assertTrue('values' in data.pop('sub'))
                self.assertTrue('values' in data.pop('txt'))
                # these are stored as singular 'value'
                self.assertTrue('value' in data.pop('aaaa'))
                self.assertTrue('value' in data.pop('cname'))
                self.assertTrue('value' in data.pop('included'))
                self.assertTrue('value' in data.pop('ptr'))
                self.assertTrue('value' in data.pop('spf'))
                self.assertTrue('value' in data.pop('www'))
                self.assertTrue('value' in data.pop('www.sub'))

                # make sure nothing is left
                self.assertEquals([], data.keys())

            with open(dynamic_yaml_file) as fh:
                data = safe_load(fh.read())

                # make sure new dynamic records made the trip
                dyna = data.pop('a')
                self.assertTrue('values' in dyna)
                # self.assertTrue('dynamic' in dyna)
                # TODO:

                # make sure new dynamic records made the trip
                dyna = data.pop('aaaa')
                self.assertTrue('values' in dyna)
                # self.assertTrue('dynamic' in dyna)

                dyna = data.pop('cname')
                self.assertTrue('value' in dyna)
                # self.assertTrue('dynamic' in dyna)

                dyna = data.pop('real-ish-a')
                self.assertTrue('values' in dyna)
                # self.assertTrue('dynamic' in dyna)

                dyna = data.pop('simple-weighted')
                self.assertTrue('value' in dyna)
                # self.assertTrue('dynamic' in dyna)

                # make sure nothing is left
                self.assertEquals([], data.keys())

    def test_empty(self):
        source = YamlProvider('test', join(dirname(__file__), 'config'))

        zone = Zone('empty.', [])

        # without it we see everything
        source.populate(zone)
        self.assertEquals(0, len(zone.records))

    def test_unsorted(self):
        source = YamlProvider('test', join(dirname(__file__), 'config'))

        zone = Zone('unordered.', [])

        with self.assertRaises(ConstructorError):
            source.populate(zone)

        source = YamlProvider('test', join(dirname(__file__), 'config'),
                              enforce_order=False)
        # no exception
        source.populate(zone)
        self.assertEqual(2, len(zone.records))

    def test_subzone_handling(self):
        source = YamlProvider('test', join(dirname(__file__), 'config'))

        # If we add `sub` as a sub-zone we'll reject `www.sub`
        zone = Zone('unit.tests.', ['sub'])
        with self.assertRaises(SubzoneRecordException) as ctx:
            source.populate(zone)
        self.assertEquals('Record www.sub.unit.tests. is under a managed '
                          'subzone', ctx.exception.message)


class TestSplitYamlProvider(TestCase):

    def test_list_all_yaml_files(self):
        yaml_files = ('foo.yaml', '1.yaml', '$unit.tests.yaml')
        all_files = ('something', 'else', '1', '$$', '-f') + yaml_files
        all_dirs = ('dir1', 'dir2/sub', 'tricky.yaml')

        with TemporaryDirectory() as td:
            directory = join(td.dirname)

            # Create some files, some of them with a .yaml extension, all of
            # them empty.
            for emptyfile in all_files:
                open(join(directory, emptyfile), 'w').close()
            # Do the same for some fake directories
            for emptydir in all_dirs:
                makedirs(join(directory, emptydir))

            # This isn't great, but given the variable nature of the temp dir
            # names, it's necessary.
            self.assertItemsEqual(
                yaml_files,
                (basename(f) for f in _list_all_yaml_files(directory)))

    def test_zone_directory(self):
        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'))

        zone = Zone('unit.tests.', [])

        self.assertEqual(
            join(dirname(__file__), 'config/split/unit.tests.'),
            source._zone_directory(zone))

    def test_apply_handles_existing_zone_directory(self):
        with TemporaryDirectory() as td:
            provider = SplitYamlProvider('test', join(td.dirname, 'config'))
            makedirs(join(td.dirname, 'config', 'does.exist.'))

            zone = Zone('does.exist.', [])
            self.assertTrue(isdir(provider._zone_directory(zone)))
            provider.apply(Plan(None, zone, [], True))
            self.assertTrue(isdir(provider._zone_directory(zone)))

    def test_provider(self):
        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'))

        zone = Zone('unit.tests.', [])
        dynamic_zone = Zone('dynamic.tests.', [])

        # With target we don't add anything
        source.populate(zone, target=source)
        self.assertEquals(0, len(zone.records))

        # without it we see everything
        source.populate(zone)
        self.assertEquals(18, len(zone.records))

        source.populate(dynamic_zone)
        self.assertEquals(5, len(dynamic_zone.records))

        with TemporaryDirectory() as td:
            # Add some subdirs to make sure that it can create them
            directory = join(td.dirname, 'sub', 'dir')
            zone_dir = join(directory, 'unit.tests.')
            dynamic_zone_dir = join(directory, 'dynamic.tests.')
            target = SplitYamlProvider('test', directory)

            # We add everything
            plan = target.plan(zone)
            self.assertEquals(15, len(filter(lambda c: isinstance(c, Create),
                                             plan.changes)))
            self.assertFalse(isdir(zone_dir))

            # Now actually do it
            self.assertEquals(15, target.apply(plan))

            # Dynamic plan
            plan = target.plan(dynamic_zone)
            self.assertEquals(5, len(filter(lambda c: isinstance(c, Create),
                                            plan.changes)))
            self.assertFalse(isdir(dynamic_zone_dir))
            # Apply it
            self.assertEquals(5, target.apply(plan))
            self.assertTrue(isdir(dynamic_zone_dir))

            # There should be no changes after the round trip
            reloaded = Zone('unit.tests.', [])
            target.populate(reloaded)
            self.assertDictEqual(
                {'included': ['test']},
                filter(
                    lambda x: x.name == 'included', reloaded.records
                )[0]._octodns)

            self.assertFalse(zone.changes(reloaded, target=source))

            # A 2nd sync should still create everything
            plan = target.plan(zone)
            self.assertEquals(15, len(filter(lambda c: isinstance(c, Create),
                                             plan.changes)))

            yaml_file = join(zone_dir, '$unit.tests.yaml')
            self.assertTrue(isfile(yaml_file))
            with open(yaml_file) as fh:
                data = safe_load(fh.read())
                roots = sorted(data.pop(''), key=lambda r: r['type'])
                self.assertTrue('values' in roots[0])  # A
                self.assertTrue('geo' in roots[0])  # geo made the trip
                self.assertTrue('value' in roots[1])   # CAA
                self.assertTrue('values' in roots[2])  # SSHFP

            # These records are stored as plural "values." Check each file to
            # ensure correctness.
            for record_name in ('_srv._tcp', 'mx', 'naptr', 'sub', 'txt'):
                yaml_file = join(zone_dir, '{}.yaml'.format(record_name))
                self.assertTrue(isfile(yaml_file))
                with open(yaml_file) as fh:
                    data = safe_load(fh.read())
                    self.assertTrue('values' in data.pop(record_name))

            # These are stored as singular "value." Again, check each file.
            for record_name in ('aaaa', 'cname', 'included', 'ptr', 'spf',
                                'www.sub', 'www'):
                yaml_file = join(zone_dir, '{}.yaml'.format(record_name))
                self.assertTrue(isfile(yaml_file))
                with open(yaml_file) as fh:
                    data = safe_load(fh.read())
                    self.assertTrue('value' in data.pop(record_name))

            # Again with the plural, this time checking dynamic.tests.
            for record_name in ('a', 'aaaa', 'real-ish-a'):
                yaml_file = join(
                    dynamic_zone_dir, '{}.yaml'.format(record_name))
                self.assertTrue(isfile(yaml_file))
                with open(yaml_file) as fh:
                    data = safe_load(fh.read())
                    dyna = data.pop(record_name)
                    self.assertTrue('values' in dyna)
                    self.assertTrue('dynamic' in dyna)

            # Singular again.
            for record_name in ('cname', 'simple-weighted'):
                yaml_file = join(
                    dynamic_zone_dir, '{}.yaml'.format(record_name))
                self.assertTrue(isfile(yaml_file))
                with open(yaml_file) as fh:
                    data = safe_load(fh.read())
                    dyna = data.pop(record_name)
                    self.assertTrue('value' in dyna)
                    self.assertTrue('dynamic' in dyna)

    def test_empty(self):
        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'))

        zone = Zone('empty.', [])

        # without it we see everything
        source.populate(zone)
        self.assertEquals(0, len(zone.records))

    def test_unsorted(self):
        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'))

        zone = Zone('unordered.', [])

        with self.assertRaises(ConstructorError):
            source.populate(zone)

        zone = Zone('unordered.', [])

        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'),
            enforce_order=False)
        # no exception
        source.populate(zone)
        self.assertEqual(2, len(zone.records))

    def test_subzone_handling(self):
        source = SplitYamlProvider(
            'test', join(dirname(__file__), 'config/split'))

        # If we add `sub` as a sub-zone we'll reject `www.sub`
        zone = Zone('unit.tests.', ['sub'])
        with self.assertRaises(SubzoneRecordException) as ctx:
            source.populate(zone)
        self.assertEquals('Record www.sub.unit.tests. is under a managed '
                          'subzone', ctx.exception.message)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from unittest import TestCase

from mock import patch, call
from ovh import APIError, ResourceNotFoundError, InvalidCredential

from octodns.provider.ovh import OvhProvider
from octodns.record import Record
from octodns.zone import Zone


class TestOvhProvider(TestCase):
    api_record = []
    valid_dkim = []
    invalid_dkim = []

    valid_dkim_key = "p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCxLaG16G4SaE" \
                     "cXVdiIxTg7gKSGbHKQLm30CHib1h9FzS9nkcyvQSyQj1rMFyqC//" \
                     "tft3ohx3nvJl+bGCWxdtLYDSmir9PW54e5CTdxEh8MWRkBO3StF6" \
                     "QG/tAh3aTGDmkqhIJGLb87iHvpmVKqURmEUzJPv5KPJfWLofADI+" \
                     "q9lQIDAQAB"

    zone = Zone('unit.tests.', [])
    expected = set()

    # A, subdomain=''
    api_record.append({
        'fieldType': 'A',
        'ttl': 100,
        'target': '1.2.3.4',
        'subDomain': '',
        'id': 1
    })
    expected.add(Record.new(zone, '', {
        'ttl': 100,
        'type': 'A',
        'value': '1.2.3.4',
    }))

    # A, subdomain='sub
    api_record.append({
        'fieldType': 'A',
        'ttl': 200,
        'target': '1.2.3.4',
        'subDomain': 'sub',
        'id': 2
    })
    expected.add(Record.new(zone, 'sub', {
        'ttl': 200,
        'type': 'A',
        'value': '1.2.3.4',
    }))

    # CNAME
    api_record.append({
        'fieldType': 'CNAME',
        'ttl': 300,
        'target': 'unit.tests.',
        'subDomain': 'www2',
        'id': 3
    })
    expected.add(Record.new(zone, 'www2', {
        'ttl': 300,
        'type': 'CNAME',
        'value': 'unit.tests.',
    }))

    # MX
    api_record.append({
        'fieldType': 'MX',
        'ttl': 400,
        'target': '10 mx1.unit.tests.',
        'subDomain': '',
        'id': 4
    })
    expected.add(Record.new(zone, '', {
        'ttl': 400,
        'type': 'MX',
        'values': [{
            'preference': 10,
            'exchange': 'mx1.unit.tests.',
        }]
    }))

    # NAPTR
    api_record.append({
        'fieldType': 'NAPTR',
        'ttl': 500,
        'target': '10 100 "S" "SIP+D2U" "!^.*$!sip:info@bar.example.com!" .',
        'subDomain': 'naptr',
        'id': 5
    })
    expected.add(Record.new(zone, 'naptr', {
        'ttl': 500,
        'type': 'NAPTR',
        'values': [{
            'flags': 'S',
            'order': 10,
            'preference': 100,
            'regexp': '!^.*$!sip:info@bar.example.com!',
            'replacement': '.',
            'service': 'SIP+D2U',
        }]
    }))

    # NS
    api_record.append({
        'fieldType': 'NS',
        'ttl': 600,
        'target': 'ns1.unit.tests.',
        'subDomain': '',
        'id': 6
    })
    api_record.append({
        'fieldType': 'NS',
        'ttl': 600,
        'target': 'ns2.unit.tests.',
        'subDomain': '',
        'id': 7
    })
    expected.add(Record.new(zone, '', {
        'ttl': 600,
        'type': 'NS',
        'values': ['ns1.unit.tests.', 'ns2.unit.tests.'],
    }))

    # NS with sub
    api_record.append({
        'fieldType': 'NS',
        'ttl': 700,
        'target': 'ns3.unit.tests.',
        'subDomain': 'www3',
        'id': 8
    })
    api_record.append({
        'fieldType': 'NS',
        'ttl': 700,
        'target': 'ns4.unit.tests.',
        'subDomain': 'www3',
        'id': 9
    })
    expected.add(Record.new(zone, 'www3', {
        'ttl': 700,
        'type': 'NS',
        'values': ['ns3.unit.tests.', 'ns4.unit.tests.'],
    }))

    api_record.append({
        'fieldType': 'SRV',
        'ttl': 800,
        'target': '10 20 30 foo-1.unit.tests.',
        'subDomain': '_srv._tcp',
        'id': 10
    })
    api_record.append({
        'fieldType': 'SRV',
        'ttl': 800,
        'target': '40 50 60 foo-2.unit.tests.',
        'subDomain': '_srv._tcp',
        'id': 11
    })
    expected.add(Record.new(zone, '_srv._tcp', {
        'ttl': 800,
        'type': 'SRV',
        'values': [{
            'priority': 10,
            'weight': 20,
            'port': 30,
            'target': 'foo-1.unit.tests.',
        }, {
            'priority': 40,
            'weight': 50,
            'port': 60,
            'target': 'foo-2.unit.tests.',
        }]
    }))

    # PTR
    api_record.append({
        'fieldType': 'PTR',
        'ttl': 900,
        'target': 'unit.tests.',
        'subDomain': '4',
        'id': 12
    })
    expected.add(Record.new(zone, '4', {
        'ttl': 900,
        'type': 'PTR',
        'value': 'unit.tests.'
    }))

    # SPF
    api_record.append({
        'fieldType': 'SPF',
        'ttl': 1000,
        'target': 'v=spf1 include:unit.texts.redirect ~all',
        'subDomain': '',
        'id': 13
    })
    expected.add(Record.new(zone, '', {
        'ttl': 1000,
        'type': 'SPF',
        'value': 'v=spf1 include:unit.texts.redirect ~all'
    }))

    # SSHFP
    api_record.append({
        'fieldType': 'SSHFP',
        'ttl': 1100,
        'target': '1 1 bf6b6825d2977c511a475bbefb88aad54a92ac73 ',
        'subDomain': '',
        'id': 14
    })
    expected.add(Record.new(zone, '', {
        'ttl': 1100,
        'type': 'SSHFP',
        'value': {
            'algorithm': 1,
            'fingerprint': 'bf6b6825d2977c511a475bbefb88aad54a92ac73',
            'fingerprint_type': 1
        }
    }))

    # AAAA
    api_record.append({
        'fieldType': 'AAAA',
        'ttl': 1200,
        'target': '1:1ec:1::1',
        'subDomain': '',
        'id': 15
    })
    expected.add(Record.new(zone, '', {
        'ttl': 200,
        'type': 'AAAA',
        'value': '1:1ec:1::1',
    }))

    # DKIM
    api_record.append({
        'fieldType': 'DKIM',
        'ttl': 1300,
        'target': valid_dkim_key,
        'subDomain': 'dkim',
        'id': 16
    })
    expected.add(Record.new(zone, 'dkim', {
        'ttl': 1300,
        'type': 'TXT',
        'value': valid_dkim_key,
    }))

    # TXT
    api_record.append({
        'fieldType': 'TXT',
        'ttl': 1400,
        'target': 'TXT text',
        'subDomain': 'txt',
        'id': 17
    })
    expected.add(Record.new(zone, 'txt', {
        'ttl': 1400,
        'type': 'TXT',
        'value': 'TXT text',
    }))

    # LOC
    # We do not have associated record for LOC, as it's not managed
    api_record.append({
        'fieldType': 'LOC',
        'ttl': 1500,
        'target': '1 1 1 N 1 1 1 E 1m 1m',
        'subDomain': '',
        'id': 18
    })

    valid_dkim = [valid_dkim_key,
                  'v=DKIM1 \\; %s' % valid_dkim_key,
                  'h=sha256 \\; %s' % valid_dkim_key,
                  'h=sha1 \\; %s' % valid_dkim_key,
                  's=* \\; %s' % valid_dkim_key,
                  's=email \\; %s' % valid_dkim_key,
                  't=y \\; %s' % valid_dkim_key,
                  't=s \\; %s' % valid_dkim_key,
                  'k=rsa \\; %s' % valid_dkim_key,
                  'n=notes \\; %s' % valid_dkim_key,
                  'g=granularity \\; %s' % valid_dkim_key,
                  ]
    invalid_dkim = ['p=%invalid%',  # Invalid public key
                    'v=DKIM1',  # Missing public key
                    'v=DKIM2 \\; %s' % valid_dkim_key,  # Invalid version
                    'h=sha512 \\; %s' % valid_dkim_key,  # Invalid hash algo
                    's=fake \\; %s' % valid_dkim_key,  # Invalid selector
                    't=fake \\; %s' % valid_dkim_key,  # Invalid flag
                    'u=invalid \\; %s' % valid_dkim_key,  # Invalid key
                    ]

    @patch('ovh.Client')
    def test_populate(self, client_mock):
        provider = OvhProvider('test', 'endpoint', 'application_key',
                               'application_secret', 'consumer_key')

        with patch.object(provider._client, 'get') as get_mock:
            zone = Zone('unit.tests.', [])
            get_mock.side_effect = ResourceNotFoundError('boom')
            with self.assertRaises(APIError) as ctx:
                provider.populate(zone)
            self.assertEquals(get_mock.side_effect, ctx.exception)

            get_mock.side_effect = InvalidCredential('boom')
            with self.assertRaises(APIError) as ctx:
                provider.populate(zone)
            self.assertEquals(get_mock.side_effect, ctx.exception)

            zone = Zone('unit.tests.', [])
            get_mock.side_effect = ResourceNotFoundError('This service does '
                                                         'not exist')
            exists = provider.populate(zone)
            self.assertEquals(set(), zone.records)
            self.assertFalse(exists)

            zone = Zone('unit.tests.', [])
            get_returns = [[record['id'] for record in self.api_record]]
            get_returns += self.api_record
            get_mock.side_effect = get_returns
            exists = provider.populate(zone)
            self.assertEquals(self.expected, zone.records)
            self.assertTrue(exists)

    @patch('ovh.Client')
    def test_is_valid_dkim(self, client_mock):
        """Test _is_valid_dkim"""
        provider = OvhProvider('test', 'endpoint', 'application_key',
                               'application_secret', 'consumer_key')
        for dkim in self.valid_dkim:
            self.assertTrue(provider._is_valid_dkim(dkim))
        for dkim in self.invalid_dkim:
            self.assertFalse(provider._is_valid_dkim(dkim))

    @patch('ovh.Client')
    def test_apply(self, client_mock):
        provider = OvhProvider('test', 'endpoint', 'application_key',
                               'application_secret', 'consumer_key')

        desired = Zone('unit.tests.', [])

        for r in self.expected:
            desired.add_record(r)

        with patch.object(provider._client, 'post') as get_mock:
            plan = provider.plan(desired)
            get_mock.side_effect = APIError('boom')
            with self.assertRaises(APIError) as ctx:
                provider.apply(plan)
            self.assertEquals(get_mock.side_effect, ctx.exception)

        # Records get by API call
        with patch.object(provider._client, 'get') as get_mock:
            get_returns = [
                [1, 2, 3, 4],
                {'fieldType': 'A', 'ttl': 600, 'target': '5.6.7.8',
                 'subDomain': '', 'id': 100},
                {'fieldType': 'A', 'ttl': 600, 'target': '5.6.7.8',
                 'subDomain': 'fake', 'id': 101},
                {'fieldType': 'TXT', 'ttl': 600, 'target': 'fake txt record',
                 'subDomain': 'txt', 'id': 102},
                {'fieldType': 'DKIM', 'ttl': 600,
                 'target': 'v=DKIM1; %s' % self.valid_dkim_key,
                 'subDomain': 'dkim', 'id': 103}
            ]
            get_mock.side_effect = get_returns

            plan = provider.plan(desired)

            with patch.object(provider._client, 'post') as post_mock, \
                    patch.object(provider._client, 'delete') as delete_mock:
                get_mock.side_effect = [[100], [101], [102], [103]]
                provider.apply(plan)
                wanted_calls = [
                    call(u'/domain/zone/unit.tests/record', fieldType=u'TXT',
                         subDomain='txt', target=u'TXT text', ttl=1400),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'DKIM',
                         subDomain='dkim', target=self.valid_dkim_key,
                         ttl=1300),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'A',
                         subDomain=u'', target=u'1.2.3.4', ttl=100),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'SRV',
                         subDomain='_srv._tcp',
                         target=u'10 20 30 foo-1.unit.tests.', ttl=800),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'SRV',
                         subDomain='_srv._tcp',
                         target=u'40 50 60 foo-2.unit.tests.', ttl=800),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'PTR',
                         subDomain='4', target=u'unit.tests.', ttl=900),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'NS',
                         subDomain='www3', target=u'ns3.unit.tests.', ttl=700),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'NS',
                         subDomain='www3', target=u'ns4.unit.tests.', ttl=700),
                    call(u'/domain/zone/unit.tests/record',
                         fieldType=u'SSHFP', subDomain=u'', ttl=1100,
                         target=u'1 1 bf6b6825d2977c511a475bbefb88a'
                                   u'ad54'
                                   u'a92ac73',
                         ),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'AAAA',
                         subDomain=u'', target=u'1:1ec:1::1', ttl=200),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'MX',
                         subDomain=u'', target=u'10 mx1.unit.tests.', ttl=400),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'CNAME',
                         subDomain='www2', target=u'unit.tests.', ttl=300),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'SPF',
                         subDomain=u'', ttl=1000,
                         target=u'v=spf1 include:unit.texts.'
                                u'redirect ~all',
                         ),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'A',
                         subDomain='sub', target=u'1.2.3.4', ttl=200),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'NAPTR',
                         subDomain='naptr', ttl=500,
                         target=u'10 100 "S" "SIP+D2U" "!^.*$!sip:'
                                u'info@bar'
                                u'.example.com!" .'
                         ),
                    call(u'/domain/zone/unit.tests/refresh')]

                post_mock.assert_has_calls(wanted_calls)

                # Get for delete calls
                wanted_get_calls = [
                    call(u'/domain/zone/unit.tests/record', fieldType=u'TXT',
                         subDomain='txt'),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'DKIM',
                         subDomain='dkim'),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'A',
                         subDomain=u''),
                    call(u'/domain/zone/unit.tests/record', fieldType=u'A',
                         subDomain='fake')]
                get_mock.assert_has_calls(wanted_get_calls)
                # 4 delete calls for update and delete
                delete_mock.assert_has_calls(
                    [call(u'/domain/zone/unit.tests/record/100'),
                     call(u'/domain/zone/unit.tests/record/101'),
                     call(u'/domain/zone/unit.tests/record/102'),
                     call(u'/domain/zone/unit.tests/record/103')])



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from mock import Mock, call, patch
from nsone.rest.errors import AuthException, RateLimitException, \
    ResourceException
from unittest import TestCase

from octodns.record import Delete, Record, Update
from octodns.provider.ns1 import Ns1Provider
from octodns.zone import Zone


class DummyZone(object):

    def __init__(self, records):
        self.data = {
            'records': records
        }


class TestNs1Provider(TestCase):
    zone = Zone('unit.tests.', [])
    expected = set()
    expected.add(Record.new(zone, '', {
        'ttl': 32,
        'type': 'A',
        'value': '1.2.3.4',
        'meta': {},
    }))
    expected.add(Record.new(zone, 'foo', {
        'ttl': 33,
        'type': 'A',
        'values': ['1.2.3.4', '1.2.3.5'],
        'meta': {},
    }))
    expected.add(Record.new(zone, 'geo', {
        'ttl': 34,
        'type': 'A',
        'values': ['101.102.103.104', '101.102.103.105'],
        'geo': {'NA-US-NY': ['201.202.203.204']},
        'meta': {},
    }))
    expected.add(Record.new(zone, 'cname', {
        'ttl': 34,
        'type': 'CNAME',
        'value': 'foo.unit.tests.',
    }))
    expected.add(Record.new(zone, '', {
        'ttl': 35,
        'type': 'MX',
        'values': [{
            'preference': 10,
            'exchange': 'mx1.unit.tests.',
        }, {
            'preference': 20,
            'exchange': 'mx2.unit.tests.',
        }]
    }))
    expected.add(Record.new(zone, 'naptr', {
        'ttl': 36,
        'type': 'NAPTR',
        'values': [{
            'flags': 'U',
            'order': 100,
            'preference': 100,
            'regexp': '!^.*$!sip:info@bar.example.com!',
            'replacement': '.',
            'service': 'SIP+D2U',
        }, {
            'flags': 'S',
            'order': 10,
            'preference': 100,
            'regexp': '!^.*$!sip:info@bar.example.com!',
            'replacement': '.',
            'service': 'SIP+D2U',
        }]
    }))
    expected.add(Record.new(zone, '', {
        'ttl': 37,
        'type': 'NS',
        'values': ['ns1.unit.tests.', 'ns2.unit.tests.'],
    }))
    expected.add(Record.new(zone, '_srv._tcp', {
        'ttl': 38,
        'type': 'SRV',
        'values': [{
            'priority': 10,
            'weight': 20,
            'port': 30,
            'target': 'foo-1.unit.tests.',
        }, {
            'priority': 12,
            'weight': 30,
            'port': 30,
            'target': 'foo-2.unit.tests.',
        }]
    }))
    expected.add(Record.new(zone, 'sub', {
        'ttl': 39,
        'type': 'NS',
        'values': ['ns3.unit.tests.', 'ns4.unit.tests.'],
    }))
    expected.add(Record.new(zone, '', {
        'ttl': 40,
        'type': 'CAA',
        'value': {
            'flags': 0,
            'tag': 'issue',
            'value': 'ca.unit.tests',
        },
    }))

    nsone_records = [{
        'type': 'A',
        'ttl': 32,
        'short_answers': ['1.2.3.4'],
        'domain': 'unit.tests.',
    }, {
        'type': 'A',
        'ttl': 33,
        'short_answers': ['1.2.3.4', '1.2.3.5'],
        'domain': 'foo.unit.tests.',
    }, {
        'type': 'A',
        'ttl': 34,
        'short_answers': ['101.102.103.104', '101.102.103.105'],
        'domain': 'geo.unit.tests',
    }, {
        'type': 'CNAME',
        'ttl': 34,
        'short_answers': ['foo.unit.tests'],
        'domain': 'cname.unit.tests.',
    }, {
        'type': 'MX',
        'ttl': 35,
        'short_answers': ['10 mx1.unit.tests.', '20 mx2.unit.tests'],
        'domain': 'unit.tests.',
    }, {
        'type': 'NAPTR',
        'ttl': 36,
        'short_answers': [
            '10 100 S SIP+D2U !^.*$!sip:info@bar.example.com! .',
            '100 100 U SIP+D2U !^.*$!sip:info@bar.example.com! .'
        ],
        'domain': 'naptr.unit.tests.',
    }, {
        'type': 'NS',
        'ttl': 37,
        'short_answers': ['ns1.unit.tests.', 'ns2.unit.tests'],
        'domain': 'unit.tests.',
    }, {
        'type': 'SRV',
        'ttl': 38,
        'short_answers': ['12 30 30 foo-2.unit.tests.',
                          '10 20 30 foo-1.unit.tests'],
        'domain': '_srv._tcp.unit.tests.',
    }, {
        'type': 'NS',
        'ttl': 39,
        'short_answers': ['ns3.unit.tests.', 'ns4.unit.tests'],
        'domain': 'sub.unit.tests.',
    }, {
        'type': 'CAA',
        'ttl': 40,
        'short_answers': ['0 issue ca.unit.tests'],
        'domain': 'unit.tests.',
    }]

    @patch('nsone.NSONE.loadZone')
    def test_populate(self, load_mock):
        provider = Ns1Provider('test', 'api-key')

        # Bad auth
        load_mock.side_effect = AuthException('unauthorized')
        zone = Zone('unit.tests.', [])
        with self.assertRaises(AuthException) as ctx:
            provider.populate(zone)
        self.assertEquals(load_mock.side_effect, ctx.exception)

        # General error
        load_mock.reset_mock()
        load_mock.side_effect = ResourceException('boom')
        zone = Zone('unit.tests.', [])
        with self.assertRaises(ResourceException) as ctx:
            provider.populate(zone)
        self.assertEquals(load_mock.side_effect, ctx.exception)
        self.assertEquals(('unit.tests',), load_mock.call_args[0])

        # Non-existant zone doesn't populate anything
        load_mock.reset_mock()
        load_mock.side_effect = \
            ResourceException('server error: zone not found')
        zone = Zone('unit.tests.', [])
        exists = provider.populate(zone)
        self.assertEquals(set(), zone.records)
        self.assertEquals(('unit.tests',), load_mock.call_args[0])
        self.assertFalse(exists)

        # Existing zone w/o records
        load_mock.reset_mock()
        nsone_zone = DummyZone([])
        load_mock.side_effect = [nsone_zone]
        zone_search = Mock()
        zone_search.return_value = [
            {
                "domain": "geo.unit.tests",
                "zone": "unit.tests",
                "type": "A",
                "answers": [
                    {'answer': ['1.1.1.1'], 'meta': {}},
                    {'answer': ['1.2.3.4'],
                     'meta': {'ca_province': ['ON']}},
                    {'answer': ['2.3.4.5'], 'meta': {'us_state': ['NY']}},
                    {'answer': ['3.4.5.6'], 'meta': {'country': ['US']}},
                    {'answer': ['4.5.6.7'],
                     'meta': {'iso_region_code': ['NA-US-WA']}},
                ],
                'ttl': 34,
            },
        ]
        nsone_zone.search = zone_search
        zone = Zone('unit.tests.', [])
        provider.populate(zone)
        self.assertEquals(1, len(zone.records))
        self.assertEquals(('unit.tests',), load_mock.call_args[0])

        # Existing zone w/records
        load_mock.reset_mock()
        nsone_zone = DummyZone(self.nsone_records)
        load_mock.side_effect = [nsone_zone]
        zone_search = Mock()
        zone_search.return_value = [
            {
                "domain": "geo.unit.tests",
                "zone": "unit.tests",
                "type": "A",
                "answers": [
                    {'answer': ['1.1.1.1'], 'meta': {}},
                    {'answer': ['1.2.3.4'],
                     'meta': {'ca_province': ['ON']}},
                    {'answer': ['2.3.4.5'], 'meta': {'us_state': ['NY']}},
                    {'answer': ['3.4.5.6'], 'meta': {'country': ['US']}},
                    {'answer': ['4.5.6.7'],
                     'meta': {'iso_region_code': ['NA-US-WA']}},
                ],
                'ttl': 34,
            },
        ]
        nsone_zone.search = zone_search
        zone = Zone('unit.tests.', [])
        provider.populate(zone)
        self.assertEquals(self.expected, zone.records)
        self.assertEquals(('unit.tests',), load_mock.call_args[0])

        # Test skipping unsupported record type
        load_mock.reset_mock()
        nsone_zone = DummyZone(self.nsone_records + [{
            'type': 'UNSUPPORTED',
            'ttl': 42,
            'short_answers': ['unsupported'],
            'domain': 'unsupported.unit.tests.',
        }])
        load_mock.side_effect = [nsone_zone]
        zone_search = Mock()
        zone_search.return_value = [
            {
                "domain": "geo.unit.tests",
                "zone": "unit.tests",
                "type": "A",
                "answers": [
                    {'answer': ['1.1.1.1'], 'meta': {}},
                    {'answer': ['1.2.3.4'],
                     'meta': {'ca_province': ['ON']}},
                    {'answer': ['2.3.4.5'], 'meta': {'us_state': ['NY']}},
                    {'answer': ['3.4.5.6'], 'meta': {'country': ['US']}},
                    {'answer': ['4.5.6.7'],
                     'meta': {'iso_region_code': ['NA-US-WA']}},
                ],
                'ttl': 34,
            },
        ]
        nsone_zone.search = zone_search
        zone = Zone('unit.tests.', [])
        provider.populate(zone)
        self.assertEquals(self.expected, zone.records)
        self.assertEquals(('unit.tests',), load_mock.call_args[0])

    @patch('nsone.NSONE.createZone')
    @patch('nsone.NSONE.loadZone')
    def test_sync(self, load_mock, create_mock):
        provider = Ns1Provider('test', 'api-key')

        desired = Zone('unit.tests.', [])
        for r in self.expected:
            desired.add_record(r)

        plan = provider.plan(desired)
        # everything except the root NS
        expected_n = len(self.expected) - 1
        self.assertEquals(expected_n, len(plan.changes))
        self.assertTrue(plan.exists)

        # Fails, general error
        load_mock.reset_mock()
        create_mock.reset_mock()
        load_mock.side_effect = ResourceException('boom')
        with self.assertRaises(ResourceException) as ctx:
            provider.apply(plan)
        self.assertEquals(load_mock.side_effect, ctx.exception)

        # Fails, bad auth
        load_mock.reset_mock()
        create_mock.reset_mock()
        load_mock.side_effect = \
            ResourceException('server error: zone not found')
        create_mock.side_effect = AuthException('unauthorized')
        with self.assertRaises(AuthException) as ctx:
            provider.apply(plan)
        self.assertEquals(create_mock.side_effect, ctx.exception)

        # non-existant zone, create
        load_mock.reset_mock()
        create_mock.reset_mock()
        load_mock.side_effect = \
            ResourceException('server error: zone not found')
        # ugh, need a mock zone with a mock prop since we're using getattr, we
        # can actually control side effects on `meth` with that.
        mock_zone = Mock()
        mock_zone.add_SRV = Mock()
        mock_zone.add_SRV.side_effect = [
            RateLimitException('boo', period=0),
            None,
        ]
        create_mock.side_effect = [mock_zone]
        got_n = provider.apply(plan)
        self.assertEquals(expected_n, got_n)

        # Update & delete
        load_mock.reset_mock()
        create_mock.reset_mock()
        nsone_zone = DummyZone(self.nsone_records + [{
            'type': 'A',
            'ttl': 42,
            'short_answers': ['9.9.9.9'],
            'domain': 'delete-me.unit.tests.',
        }])
        nsone_zone.data['records'][0]['short_answers'][0] = '2.2.2.2'
        nsone_zone.loadRecord = Mock()
        zone_search = Mock()
        zone_search.return_value = [
            {
                "domain": "geo.unit.tests",
                "zone": "unit.tests",
                "type": "A",
                "answers": [
                    {'answer': ['1.1.1.1'], 'meta': {}},
                    {'answer': ['1.2.3.4'],
                     'meta': {'ca_province': ['ON']}},
                    {'answer': ['2.3.4.5'], 'meta': {'us_state': ['NY']}},
                    {'answer': ['3.4.5.6'], 'meta': {'country': ['US']}},
                    {'answer': ['4.5.6.7'],
                     'meta': {'iso_region_code': ['NA-US-WA']}},
                ],
                'ttl': 34,
            },
        ]
        nsone_zone.search = zone_search
        load_mock.side_effect = [nsone_zone, nsone_zone]
        plan = provider.plan(desired)
        self.assertEquals(3, len(plan.changes))
        self.assertIsInstance(plan.changes[0], Update)
        self.assertIsInstance(plan.changes[2], Delete)
        # ugh, we need a mock record that can be returned from loadRecord for
        # the update and delete targets, we can add our side effects to that to
        # trigger rate limit handling
        mock_record = Mock()
        mock_record.update.side_effect = [
            RateLimitException('one', period=0),
            None,
            None,
        ]
        mock_record.delete.side_effect = [
            RateLimitException('two', period=0),
            None,
            None,
        ]
        nsone_zone.loadRecord.side_effect = [mock_record, mock_record,
                                             mock_record]
        got_n = provider.apply(plan)
        self.assertEquals(3, got_n)
        nsone_zone.loadRecord.assert_has_calls([
            call('unit.tests', u'A'),
            call('geo', u'A'),
            call('delete-me', u'A'),
        ])
        mock_record.assert_has_calls([
            call.update(answers=[{'answer': [u'1.2.3.4'], 'meta': {}}],
                        filters=[],
                        ttl=32),
            call.update(answers=[{u'answer': [u'1.2.3.4'], u'meta': {}}],
                        filters=[],
                        ttl=32),
            call.update(
                answers=[
                    {u'answer': [u'101.102.103.104'], u'meta': {}},
                    {u'answer': [u'101.102.103.105'], u'meta': {}},
                    {
                        u'answer': [u'201.202.203.204'],
                        u'meta': {
                            u'iso_region_code': [u'NA-US-NY']
                        },
                    },
                ],
                filters=[
                    {u'filter': u'shuffle', u'config': {}},
                    {u'filter': u'geotarget_country', u'config': {}},
                    {u'filter': u'select_first_n', u'config': {u'N': 1}},
                ],
                ttl=34),
            call.delete(),
            call.delete()
        ])

    def test_escaping(self):
        provider = Ns1Provider('test', 'api-key')
        record = {
            'ttl': 31,
            'short_answers': ['foo; bar baz; blip']
        }
        self.assertEquals(['foo\\; bar baz\\; blip'],
                          provider._data_for_SPF('SPF', record)['values'])

        record = {
            'ttl': 31,
            'short_answers': ['no', 'foo; bar baz; blip', 'yes']
        }
        self.assertEquals(['no', 'foo\\; bar baz\\; blip', 'yes'],
                          provider._data_for_TXT('TXT', record)['values'])

        zone = Zone('unit.tests.', [])
        record = Record.new(zone, 'spf', {
            'ttl': 34,
            'type': 'SPF',
            'value': 'foo\\; bar baz\\; blip'
        })
        self.assertEquals(['foo; bar baz; blip'],
                          provider._params_for_SPF(record)['answers'])

        record = Record.new(zone, 'txt', {
            'ttl': 35,
            'type': 'TXT',
            'value': 'foo\\; bar baz\\; blip'
        })
        self.assertEquals(['foo; bar baz; blip'],
                          provider._params_for_TXT(record)['answers'])

    def test_data_for_CNAME(self):
        provider = Ns1Provider('test', 'api-key')

        # answers from nsone
        a_record = {
            'ttl': 31,
            'type': 'CNAME',
            'short_answers': ['foo.unit.tests.']
        }
        a_expected = {
            'ttl': 31,
            'type': 'CNAME',
            'value': 'foo.unit.tests.'
        }
        self.assertEqual(a_expected,
                         provider._data_for_CNAME(a_record['type'], a_record))

        # no answers from nsone
        b_record = {
            'ttl': 32,
            'type': 'CNAME',
            'short_answers': []
        }
        b_expected = {
            'ttl': 32,
            'type': 'CNAME',
            'value': None
        }
        self.assertEqual(b_expected,
                         provider._data_for_CNAME(b_record['type'], b_record))



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from requests import HTTPError, Session
import logging

from ..record import Create, Record
from .base import BaseProvider


class PowerDnsBaseProvider(BaseProvider):
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS = set(('A', 'AAAA', 'ALIAS', 'CAA', 'CNAME', 'MX', 'NAPTR', 'NS',
                    'PTR', 'SPF', 'SSHFP', 'SRV', 'TXT'))
    TIMEOUT = 5

    def __init__(self, id, host, api_key, port=8081, scheme="http",
                 timeout=TIMEOUT, *args, **kwargs):
        super(PowerDnsBaseProvider, self).__init__(id, *args, **kwargs)

        self.host = host
        self.port = port
        self.scheme = scheme
        self.timeout = timeout

        sess = Session()
        sess.headers.update({'X-API-Key': api_key})
        self._sess = sess

    def _request(self, method, path, data=None):
        self.log.debug('_request: method=%s, path=%s', method, path)

        url = '{}://{}:{}/api/v1/servers/localhost/{}' \
            .format(self.scheme, self.host, self.port, path)
        resp = self._sess.request(method, url, json=data, timeout=self.timeout)
        self.log.debug('_request:   status=%d', resp.status_code)
        resp.raise_for_status()
        return resp

    def _get(self, path, data=None):
        return self._request('GET', path, data=data)

    def _post(self, path, data=None):
        return self._request('POST', path, data=data)

    def _patch(self, path, data=None):
        return self._request('PATCH', path, data=data)

    def _data_for_multiple(self, rrset):
        # TODO: geo not supported
        return {
            'type': rrset['type'],
            'values': [r['content'] for r in rrset['records']],
            'ttl': rrset['ttl']
        }

    _data_for_A = _data_for_multiple
    _data_for_AAAA = _data_for_multiple
    _data_for_NS = _data_for_multiple

    def _data_for_CAA(self, rrset):
        values = []
        for record in rrset['records']:
            flags, tag, value = record['content'].split(' ', 2)
            values.append({
                'flags': flags,
                'tag': tag,
                'value': value[1:-1],
            })
        return {
            'type': rrset['type'],
            'values': values,
            'ttl': rrset['ttl']
        }

    def _data_for_single(self, rrset):
        return {
            'type': rrset['type'],
            'value': rrset['records'][0]['content'],
            'ttl': rrset['ttl']
        }

    _data_for_ALIAS = _data_for_single
    _data_for_CNAME = _data_for_single
    _data_for_PTR = _data_for_single

    def _data_for_quoted(self, rrset):
        return {
            'type': rrset['type'],
            'values': [r['content'][1:-1] for r in rrset['records']],
            'ttl': rrset['ttl']
        }

    _data_for_SPF = _data_for_quoted
    _data_for_TXT = _data_for_quoted

    def _data_for_MX(self, rrset):
        values = []
        for record in rrset['records']:
            preference, exchange = record['content'].split(' ', 1)
            values.append({
                'preference': preference,
                'exchange': exchange,
            })
        return {
            'type': rrset['type'],
            'values': values,
            'ttl': rrset['ttl']
        }

    def _data_for_NAPTR(self, rrset):
        values = []
        for record in rrset['records']:
            order, preference, flags, service, regexp, replacement = \
                record['content'].split(' ', 5)
            values.append({
                'order': order,
                'preference': preference,
                'flags': flags[1:-1],
                'service': service[1:-1],
                'regexp': regexp[1:-1],
                'replacement': replacement,
            })
        return {
            'type': rrset['type'],
            'values': values,
            'ttl': rrset['ttl']
        }

    def _data_for_SSHFP(self, rrset):
        values = []
        for record in rrset['records']:
            algorithm, fingerprint_type, fingerprint = \
                record['content'].split(' ', 2)
            values.append({
                'algorithm': algorithm,
                'fingerprint_type': fingerprint_type,
                'fingerprint': fingerprint,
            })
        return {
            'type': rrset['type'],
            'values': values,
            'ttl': rrset['ttl']
        }

    def _data_for_SRV(self, rrset):
        values = []
        for record in rrset['records']:
            priority, weight, port, target = \
                record['content'].split(' ', 3)
            values.append({
                'priority': priority,
                'weight': weight,
                'port': port,
                'target': target,
            })
        return {
            'type': rrset['type'],
            'values': values,
            'ttl': rrset['ttl']
        }

    def populate(self, zone, target=False, lenient=False):
        self.log.debug('populate: name=%s, target=%s, lenient=%s', zone.name,
                       target, lenient)

        resp = None
        try:
            resp = self._get('zones/{}'.format(zone.name))
            self.log.debug('populate:   loaded')
        except HTTPError as e:
            if e.response.status_code == 401:
                # Nicer error message for auth problems
                raise Exception('PowerDNS unauthorized host={}'
                                .format(self.host))
            elif e.response.status_code == 422:
                # 422 means powerdns doesn't know anything about the requested
                # domain. We'll just ignore it here and leave the zone
                # untouched.
                pass
            else:
                # just re-throw
                raise

        before = len(zone.records)
        exists = False

        if resp:
            exists = True
            for rrset in resp.json()['rrsets']:
                _type = rrset['type']
                if _type == 'SOA':
                    continue
                data_for = getattr(self, '_data_for_{}'.format(_type))
                record_name = zone.hostname_from_fqdn(rrset['name'])
                record = Record.new(zone, record_name, data_for(rrset),
                                    source=self, lenient=lenient)
                zone.add_record(record, lenient=lenient)

        self.log.info('populate:   found %s records, exists=%s',
                      len(zone.records) - before, exists)
        return exists

    def _records_for_multiple(self, record):
        return [{'content': v, 'disabled': False}
                for v in record.values]

    _records_for_A = _records_for_multiple
    _records_for_AAAA = _records_for_multiple
    _records_for_NS = _records_for_multiple

    def _records_for_CAA(self, record):
        return [{
            'content': '{} {} "{}"'.format(v.flags, v.tag, v.value),
            'disabled': False
        } for v in record.values]

    def _records_for_single(self, record):
        return [{'content': record.value, 'disabled': False}]

    _records_for_ALIAS = _records_for_single
    _records_for_CNAME = _records_for_single
    _records_for_PTR = _records_for_single

    def _records_for_quoted(self, record):
        return [{'content': '"{}"'.format(v), 'disabled': False}
                for v in record.values]

    _records_for_SPF = _records_for_quoted
    _records_for_TXT = _records_for_quoted

    def _records_for_MX(self, record):
        return [{
            'content': '{} {}'.format(v.preference, v.exchange),
            'disabled': False
        } for v in record.values]

    def _records_for_NAPTR(self, record):
        return [{
            'content': '{} {} "{}" "{}" "{}" {}'.format(v.order, v.preference,
                                                        v.flags, v.service,
                                                        v.regexp,
                                                        v.replacement),
            'disabled': False
        } for v in record.values]

    def _records_for_SSHFP(self, record):
        return [{
            'content': '{} {} {}'.format(v.algorithm, v.fingerprint_type,
                                         v.fingerprint),
            'disabled': False
        } for v in record.values]

    def _records_for_SRV(self, record):
        return [{
            'content': '{} {} {} {}'.format(v.priority, v.weight, v.port,
                                            v.target),
            'disabled': False
        } for v in record.values]

    def _mod_Create(self, change):
        new = change.new
        records_for = getattr(self, '_records_for_{}'.format(new._type))
        return {
            'name': new.fqdn,
            'type': new._type,
            'ttl': new.ttl,
            'changetype': 'REPLACE',
            'records': records_for(new)
        }

    _mod_Update = _mod_Create

    def _mod_Delete(self, change):
        existing = change.existing
        records_for = getattr(self, '_records_for_{}'.format(existing._type))
        return {
            'name': existing.fqdn,
            'type': existing._type,
            'ttl': existing.ttl,
            'changetype': 'DELETE',
            'records': records_for(existing)
        }

    def _get_nameserver_record(self, existing):
        return None

    def _extra_changes(self, existing, **kwargs):
        self.log.debug('_extra_changes: zone=%s', existing.name)

        ns = self._get_nameserver_record(existing)
        if not ns:
            return []

        # sorting mostly to make things deterministic for testing, but in
        # theory it let us find what we're after quicker (though sorting would
        # be more expensive.)
        for record in sorted(existing.records):
            if record == ns:
                # We've found the top-level NS record, return any changes
                change = record.changes(ns, self)
                self.log.debug('_extra_changes:   change=%s', change)
                if change:
                    # We need to modify an existing record
                    return [change]
                # No change is necessary
                return []
        # No existing top-level NS
        self.log.debug('_extra_changes:   create')
        return [Create(ns)]

    def _get_error(self, http_error):
        try:
            return http_error.response.json()['error']
        except Exception:
            return ''

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug('_apply: zone=%s, len(changes)=%d', desired.name,
                       len(changes))

        mods = []
        for change in changes:
            class_name = change.__class__.__name__
            mods.append(getattr(self, '_mod_{}'.format(class_name))(change))
        self.log.debug('_apply:   sending change request')

        try:
            self._patch('zones/{}'.format(desired.name),
                        data={'rrsets': mods})
            self.log.debug('_apply:   patched')
        except HTTPError as e:
            error = self._get_error(e)
            if e.response.status_code != 422 or \
               not error.startswith('Could not find domain '):
                self.log.error('_apply:   status=%d, text=%s',
                               e.response.status_code,
                               e.response.text)
                raise
            self.log.info('_apply:   creating zone=%s', desired.name)
            # 422 means powerdns doesn't know anything about the requested
            # domain. We'll try to create it with the correct records instead
            # of update. Hopefully all the mods are creates :-)
            data = {
                'name': desired.name,
                'kind': 'Master',
                'masters': [],
                'nameservers': [],
                'rrsets': mods,
                'soa_edit_api': 'INCEPTION-INCREMENT',
                'serial': 0,
            }
            try:
                self._post('zones', data)
            except HTTPError as e:
                self.log.error('_apply:   status=%d, text=%s',
                               e.response.status_code,
                               e.response.text)
                raise
            self.log.debug('_apply:   created')

        self.log.debug('_apply:   complete')


class PowerDnsProvider(PowerDnsBaseProvider):
    '''
    PowerDNS API v4 Provider

    powerdns:
        class: octodns.provider.powerdns.PowerDnsProvider
        # The host on which PowerDNS api is listening (required)
        host: fqdn
        # The api key that grans access (required)
        api_key: api-key
        # The port on which PowerDNS api is listening (optional, default 8081)
        port: 8081
        # The nameservers to use for this provider (optional,
        #   default unmanaged)
        nameserver_values:
            - 1.2.3.4.
            - 1.2.3.5.
        # The nameserver record TTL when managed, (optional, default 600)
        nameserver_ttl: 600
    '''

    def __init__(self, id, host, api_key, port=8081, nameserver_values=None,
                 nameserver_ttl=600, *args, **kwargs):
        self.log = logging.getLogger('PowerDnsProvider[{}]'.format(id))
        self.log.debug('__init__: id=%s, host=%s, port=%d, '
                       'nameserver_values=%s, nameserver_ttl=%d',
                       id, host, port, nameserver_values, nameserver_ttl)
        super(PowerDnsProvider, self).__init__(id, host=host, api_key=api_key,
                                               port=port, *args, **kwargs)

        self.nameserver_values = nameserver_values
        self.nameserver_ttl = nameserver_ttl

    def _get_nameserver_record(self, existing):
        if self.nameserver_values:
            return Record.new(existing, '', {
                'type': 'NS',
                'ttl': self.nameserver_ttl,
                'values': self.nameserver_values,
            }, source=self)

        return super(PowerDnsProvider, self)._get_nameserver_record(existing)



#
#
#
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from requests import HTTPError, Session, post
from collections import defaultdict
import logging
import string
import time

from ..record import Record
from .base import BaseProvider


def add_trailing_dot(s):
    assert s
    assert s[-1] != '.'
    return s + '.'


def remove_trailing_dot(s):
    assert s
    assert s[-1] == '.'
    return s[:-1]


def escape_semicolon(s):
    assert s
    return string.replace(s, ';', '\\;')


def unescape_semicolon(s):
    assert s
    return string.replace(s, '\\;', ';')


class RackspaceProvider(BaseProvider):
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS = set(('A', 'AAAA', 'ALIAS', 'CNAME', 'MX', 'NS', 'PTR', 'SPF',
                    'TXT'))
    TIMEOUT = 5

    def __init__(self, id, username, api_key, ratelimit_delay=0.0, *args,
                 **kwargs):
        '''
        Rackspace API v1 Provider

        rackspace:
            class: octodns.provider.rackspace.RackspaceProvider
            # The the username to authenticate with (required)
            username: username
            # The api key that grants access for that user (required)
            api_key: api-key
        '''
        self.log = logging.getLogger('RackspaceProvider[{}]'.format(id))
        super(RackspaceProvider, self).__init__(id, *args, **kwargs)

        auth_token, dns_endpoint = self._get_auth_token(username, api_key)
        self.dns_endpoint = dns_endpoint

        self.ratelimit_delay = float(ratelimit_delay)

        sess = Session()
        sess.headers.update({'X-Auth-Token': auth_token})
        self._sess = sess

        # Map record type, name, and data to an id when populating so that
        # we can find the id for update and delete operations.
        self._id_map = {}

    def _get_auth_token(self, username, api_key):
        ret = post('https://identity.api.rackspacecloud.com/v2.0/tokens',
                   json={"auth": {
                       "RAX-KSKEY:apiKeyCredentials": {"username": username,
                                                       "apiKey": api_key}}},
                   )
        cloud_dns_endpoint = \
            [x for x in ret.json()['access']['serviceCatalog'] if
             x['name'] == 'cloudDNS'][0]['endpoints'][0]['publicURL']
        return ret.json()['access']['token']['id'], cloud_dns_endpoint

    def _get_zone_id_for(self, zone):
        ret = self._request('GET', 'domains', pagination_key='domains')
        return [x for x in ret if x['name'] == zone.name[:-1]][0]['id']

    def _request(self, method, path, data=None, pagination_key=None):
        self.log.debug('_request: method=%s, path=%s', method, path)
        url = '{}/{}'.format(self.dns_endpoint, path)

        if pagination_key:
            resp = self._paginated_request_for_url(method, url, data,
                                                   pagination_key)
        else:
            resp = self._request_for_url(method, url, data)
        time.sleep(self.ratelimit_delay)
        return resp

    def _request_for_url(self, method, url, data):
        resp = self._sess.request(method, url, json=data, timeout=self.TIMEOUT)
        self.log.debug('_request:   status=%d', resp.status_code)
        resp.raise_for_status()
        return resp

    def _paginated_request_for_url(self, method, url, data, pagination_key):
        acc = []

        resp = self._sess.request(method, url, json=data, timeout=self.TIMEOUT)
        self.log.debug('_request:   status=%d', resp.status_code)
        resp.raise_for_status()
        acc.extend(resp.json()[pagination_key])

        next_page = [x for x in resp.json().get('links', []) if
                     x['rel'] == 'next']
        if next_page:
            url = next_page[0]['href']
            acc.extend(self._paginated_request_for_url(method, url, data,
                                                       pagination_key))
            return acc
        else:
            return acc

    def _post(self, path, data=None):
        return self._request('POST', path, data=data)

    def _put(self, path, data=None):
        return self._request('PUT', path, data=data)

    def _delete(self, path, data=None):
        return self._request('DELETE', path, data=data)

    @classmethod
    def _key_for_record(cls, rs_record):
        return rs_record['type'], rs_record['name'], rs_record['data']

    def _data_for_multiple(self, rrset):
        return {
            'type': rrset[0]['type'],
            'values': [r['data'] for r in rrset],
            'ttl': rrset[0]['ttl']
        }

    _data_for_A = _data_for_multiple
    _data_for_AAAA = _data_for_multiple

    def _data_for_NS(self, rrset):
        return {
            'type': rrset[0]['type'],
            'values': [add_trailing_dot(r['data']) for r in rrset],
            'ttl': rrset[0]['ttl']
        }

    def _data_for_single(self, record):
        return {
            'type': record[0]['type'],
            'value': add_trailing_dot(record[0]['data']),
            'ttl': record[0]['ttl']
        }

    _data_for_ALIAS = _data_for_single
    _data_for_CNAME = _data_for_single
    _data_for_PTR = _data_for_single

    def _data_for_textual(self, rrset):
        return {
            'type': rrset[0]['type'],
            'values': [escape_semicolon(r['data']) for r in rrset],
            'ttl': rrset[0]['ttl']
        }

    _data_for_SPF = _data_for_textual
    _data_for_TXT = _data_for_textual

    def _data_for_MX(self, rrset):
        values = []
        for record in rrset:
            values.append({
                'priority': record['priority'],
                'value': add_trailing_dot(record['data']),
            })
        return {
            'type': rrset[0]['type'],
            'values': values,
            'ttl': rrset[0]['ttl']
        }

    def populate(self, zone, target=False, lenient=False):
        self.log.debug('populate: name=%s', zone.name)
        resp_data = None
        try:
            domain_id = self._get_zone_id_for(zone)
            resp_data = self._request('GET',
                                      'domains/{}/records'.format(domain_id),
                                      pagination_key='records')
            self.log.debug('populate:   loaded')
        except HTTPError as e:
            if e.response.status_code == 401:
                # Nicer error message for auth problems
                raise Exception('Rackspace request unauthorized')
            elif e.response.status_code == 404:
                # Zone not found leaves the zone empty instead of failing.
                return False
            raise

        before = len(zone.records)

        if resp_data:
            records = self._group_records(resp_data)
            for record_type, records_of_type in records.items():
                for raw_record_name, record_set in records_of_type.items():
                    data_for = getattr(self,
                                       '_data_for_{}'.format(record_type))
                    record_name = zone.hostname_from_fqdn(raw_record_name)
                    record = Record.new(zone, record_name,
                                        data_for(record_set),
                                        source=self)
                    zone.add_record(record, lenient=lenient)

        self.log.info('populate:   found %s records, exists=True',
                      len(zone.records) - before)
        return True

    def _group_records(self, all_records):
        records = defaultdict(lambda: defaultdict(list))
        for record in all_records:
            self._id_map[self._key_for_record(record)] = record['id']
            records[record['type']][record['name']].append(record)
        return records

    @staticmethod
    def _record_for_single(record, value):
        return {
            'name': remove_trailing_dot(record.fqdn),
            'type': record._type,
            'data': value,
            'ttl': max(record.ttl, 300),
        }

    _record_for_A = _record_for_single
    _record_for_AAAA = _record_for_single

    @staticmethod
    def _record_for_named(record, value):
        return {
            'name': remove_trailing_dot(record.fqdn),
            'type': record._type,
            'data': remove_trailing_dot(value),
            'ttl': max(record.ttl, 300),
        }

    _record_for_NS = _record_for_named
    _record_for_ALIAS = _record_for_named
    _record_for_CNAME = _record_for_named
    _record_for_PTR = _record_for_named

    @staticmethod
    def _record_for_textual(record, value):
        return {
            'name': remove_trailing_dot(record.fqdn),
            'type': record._type,
            'data': unescape_semicolon(value),
            'ttl': max(record.ttl, 300),
        }

    _record_for_SPF = _record_for_textual
    _record_for_TXT = _record_for_textual

    @staticmethod
    def _record_for_MX(record, value):
        return {
            'name': remove_trailing_dot(record.fqdn),
            'type': record._type,
            'data': remove_trailing_dot(value.exchange),
            'ttl': max(record.ttl, 300),
            'priority': value.preference
        }

    def _get_values(self, record):
        try:
            return record.values
        except AttributeError:
            return [record.value]

    def _mod_Create(self, change):
        return self._create_given_change_values(change,
                                                self._get_values(change.new))

    def _create_given_change_values(self, change, values):
        transformer = getattr(self, "_record_for_{}".format(change.new._type))
        return [transformer(change.new, v) for v in values]

    def _mod_Update(self, change):
        existing_values = self._get_values(change.existing)
        new_values = self._get_values(change.new)

        # A reduction in number of values in an update record needs
        # to get upgraded into a Delete change for the removed values.
        deleted_values = set(existing_values) - set(new_values)
        delete_out = self._delete_given_change_values(change, deleted_values)

        # An increase in number of values in an update record needs
        # to get upgraded into a Create change for the added values.
        create_values = set(new_values) - set(existing_values)
        create_out = self._create_given_change_values(change, create_values)

        update_out = []
        update_values = set(new_values).intersection(set(existing_values))
        for value in update_values:
            transformer = getattr(self,
                                  "_record_for_{}".format(change.new._type))
            prior_rs_record = transformer(change.existing, value)
            prior_key = self._key_for_record(prior_rs_record)
            next_rs_record = transformer(change.new, value)
            next_key = self._key_for_record(next_rs_record)
            next_rs_record["id"] = self._id_map[prior_key]
            del next_rs_record["type"]
            update_out.append(next_rs_record)
            self._id_map[next_key] = self._id_map[prior_key]
            del self._id_map[prior_key]
        return create_out, update_out, delete_out

    def _mod_Delete(self, change):
        return self._delete_given_change_values(change, self._get_values(
            change.existing))

    def _delete_given_change_values(self, change, values):
        transformer = getattr(self, "_record_for_{}".format(
            change.existing._type))
        out = []
        for value in values:
            rs_record = transformer(change.existing, value)
            key = self._key_for_record(rs_record)
            out.append('id=' + self._id_map[key])
            del self._id_map[key]
        return out

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug('_apply: zone=%s, len(changes)=%d', desired.name,
                       len(changes))

        # Creates, updates, and deletes are processed by different endpoints
        # and are broken out by record-set entries; pre-process everything
        # into these buckets in order to minimize the number of API calls.
        domain_id = self._get_zone_id_for(desired)
        creates = []
        updates = []
        deletes = []
        for change in changes:
            if change.__class__.__name__ == 'Create':
                creates += self._mod_Create(change)
            elif change.__class__.__name__ == 'Update':
                add_creates, add_updates, add_deletes = self._mod_Update(
                    change)
                creates += add_creates
                updates += add_updates
                deletes += add_deletes
            else:
                assert change.__class__.__name__ == 'Delete'
                deletes += self._mod_Delete(change)

        if deletes:
            params = "&".join(sorted(deletes))
            self._delete('domains/{}/records?{}'.format(domain_id, params))

        if updates:
            data = {"records": sorted(updates, key=lambda v: v['name'])}
            self._put('domains/{}/records'.format(domain_id), data=data)

        if creates:
            data = {"records": sorted(creates, key=lambda v: v['type'] +
                                      v['name'] +
                                      v.get('data', ''))}
            self._post('domains/{}/records'.format(domain_id), data=data)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import defaultdict
from logging import getLogger
import re

from .record import Create, Delete


class SubzoneRecordException(Exception):
    pass


class DuplicateRecordException(Exception):
    pass


class InvalidNodeException(Exception):
    pass


def _is_eligible(record):
    # Should this record be considered when computing changes
    # We ignore all top-level NS records
    return record._type != 'NS' or record.name != ''


class Zone(object):
    log = getLogger('Zone')

    def __init__(self, name, sub_zones):
        if not name[-1] == '.':
            raise Exception('Invalid zone name {}, missing ending dot'
                            .format(name))
        # Force everything to lowercase just to be safe
        self.name = unicode(name).lower() if name else name
        self.sub_zones = sub_zones
        # We're grouping by node, it allows us to efficiently search for
        # duplicates and detect when CNAMEs co-exist with other records
        self._records = defaultdict(set)
        # optional leading . to match empty hostname
        # optional trailing . b/c some sources don't have it on their fqdn
        self._name_re = re.compile(r'\.?{}?$'.format(name))

        self.log.debug('__init__: zone=%s, sub_zones=%s', self, sub_zones)

    @property
    def records(self):
        return set([r for _, node in self._records.items() for r in node])

    def hostname_from_fqdn(self, fqdn):
        return self._name_re.sub('', fqdn)

    def add_record(self, record, replace=False, lenient=False):
        name = record.name
        last = name.split('.')[-1]

        if not lenient and last in self.sub_zones:
            if name != last:
                # it's a record for something under a sub-zone
                raise SubzoneRecordException('Record {} is under a '
                                             'managed subzone'
                                             .format(record.fqdn))
            elif record._type != 'NS':
                # It's a non NS record for exactly a sub-zone
                raise SubzoneRecordException('Record {} a managed sub-zone '
                                             'and not of type NS'
                                             .format(record.fqdn))

        if replace:
            # will remove it if it exists
            self._records[name].discard(record)

        node = self._records[name]
        if record in node:
            # We already have a record at this node of this type
            raise DuplicateRecordException('Duplicate record {}, type {}'
                                           .format(record.fqdn,
                                                   record._type))
        elif not lenient and (((record._type == 'CNAME' and len(node) > 0) or
                               ('CNAME' in map(lambda r: r._type, node)))):
            # We're adding a CNAME to existing records or adding to an existing
            # CNAME
            raise InvalidNodeException('Invalid state, CNAME at {} cannot '
                                       'coexist with other records'
                                       .format(record.fqdn))

        node.add(record)

    def _remove_record(self, record):
        'Only for use in tests'
        self._records[record.name].discard(record)

    def changes(self, desired, target):
        self.log.debug('changes: zone=%s, target=%s', self, target)

        # Build up a hash of the desired records, thanks to our special
        # __hash__ and __cmp__ on Record we'll be able to look up records that
        # match name and _type with it
        desired_records = {r: r for r in desired.records}

        changes = []

        # Find diffs & removes
        for record in filter(_is_eligible, self.records):
            if record.ignored:
                continue
            elif len(record.included) > 0 and \
                    target.id not in record.included:
                self.log.debug('changes:  skipping record=%s %s - %s not'
                               ' included ', record.fqdn, record._type,
                               target.id)
                continue
            elif target.id in record.excluded:
                self.log.debug('changes:  skipping record=%s %s - %s '
                               'excluded ', record.fqdn, record._type,
                               target.id)
                continue
            try:
                desired_record = desired_records[record]
                if desired_record.ignored:
                    continue
                elif len(desired_record.included) > 0 and \
                        target.id not in desired_record.included:
                    self.log.debug('changes:  skipping record=%s %s - %s'
                                   'not included ', record.fqdn, record._type,
                                   target.id)
                    continue
                elif target.id in desired_record.excluded:
                    continue
            except KeyError:
                if not target.supports(record):
                    self.log.debug('changes:  skipping record=%s %s - %s does '
                                   'not support it', record.fqdn, record._type,
                                   target.id)
                    continue
                # record has been removed
                self.log.debug('changes: zone=%s, removed record=%s', self,
                               record)
                changes.append(Delete(record))
            else:
                change = record.changes(desired_record, target)
                if change:
                    self.log.debug('changes: zone=%s, modified\n'
                                   '    existing=%s,\n     desired=%s', self,
                                   record, desired_record)
                    changes.append(change)
                else:
                    self.log.debug('changes: zone=%s, n.c. record=%s', self,
                                   record)

        # Find additions, things that are in desired, but missing in ourselves.
        # This uses set math and our special __hash__ and __cmp__ functions as
        # well
        for record in filter(_is_eligible, desired.records - self.records):
            if record.ignored:
                continue
            elif len(record.included) > 0 and \
                    target.id not in record.included:
                self.log.debug('changes:  skipping record=%s %s - %s not'
                               ' included ', record.fqdn, record._type,
                               target.id)
                continue
            elif target.id in record.excluded:
                self.log.debug('changes:  skipping record=%s %s - %s '
                               'excluded ', record.fqdn, record._type,
                               target.id)
                continue

            if not target.supports(record):
                self.log.debug('changes:  skipping record=%s %s - %s does not '
                               'support it', record.fqdn, record._type,
                               target.id)
                continue
            self.log.debug('changes: zone=%s, create record=%s', self, record)
            changes.append(Create(record))

        return changes

    def __repr__(self):
        return 'Zone<{}>'.format(self.name)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from ..source.base import BaseSource
from ..zone import Zone
from .plan import Plan


class BaseProvider(BaseSource):

    def __init__(self, id, apply_disabled=False,
                 update_pcent_threshold=Plan.MAX_SAFE_UPDATE_PCENT,
                 delete_pcent_threshold=Plan.MAX_SAFE_DELETE_PCENT):
        super(BaseProvider, self).__init__(id)
        self.log.debug('__init__: id=%s, apply_disabled=%s, '
                       'update_pcent_threshold=%.2f, '
                       'delete_pcent_threshold=%.2f',
                       id,
                       apply_disabled,
                       update_pcent_threshold,
                       delete_pcent_threshold)
        self.apply_disabled = apply_disabled
        self.update_pcent_threshold = update_pcent_threshold
        self.delete_pcent_threshold = delete_pcent_threshold

    def _include_change(self, change):
        '''
        An opportunity for providers to filter out false positives due to
        peculiarities in their implementation. E.g. minimum TTLs.
        '''
        return True

    def _extra_changes(self, existing, desired, changes):
        '''
        An opportunity for providers to add extra changes to the plan that are
        necessary to update ancillary record data or configure the zone. E.g.
        base NS records.
        '''
        return []

    def plan(self, desired):
        self.log.info('plan: desired=%s', desired.name)

        existing = Zone(desired.name, desired.sub_zones)
        exists = self.populate(existing, target=True, lenient=True)
        if exists is None:
            # If your code gets this warning see Source.populate for more
            # information
            self.log.warn('Provider %s used in target mode did not return '
                          'exists', self.id)

        # compute the changes at the zone/record level
        changes = existing.changes(desired, self)

        # allow the provider to filter out false positives
        before = len(changes)
        changes = filter(self._include_change, changes)
        after = len(changes)
        if before != after:
            self.log.info('plan:   filtered out %s changes', before - after)

        # allow the provider to add extra changes it needs
        extra = self._extra_changes(existing=existing, desired=desired,
                                    changes=changes)
        if extra:
            self.log.info('plan:   extra changes\n  %s', '\n  '
                          .join([unicode(c) for c in extra]))
            changes += extra

        if changes:
            plan = Plan(existing, desired, changes, exists,
                        self.update_pcent_threshold,
                        self.delete_pcent_threshold)
            self.log.info('plan:   %s', plan)
            return plan
        self.log.info('plan:   No changes')
        return None

    def apply(self, plan):
        '''
        Submits actual planned changes to the provider. Returns the number of
        changes made
        '''
        if self.apply_disabled:
            self.log.info('apply: disabled')
            return 0

        self.log.info('apply: making changes')
        self._apply(plan)
        return len(plan.changes)

    def _apply(self, plan):
        raise NotImplementedError('Abstract base class, _apply method '
                                  'missing')



#!/usr/bin/env python
'''
Octo-DNS Validator
'''

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from logging import WARN

from octodns.cmds.args import ArgumentParser
from octodns.manager import Manager


def main():
    parser = ArgumentParser(description=__doc__.split('\n')[1])

    parser.add_argument('--config-file', required=True,
                        help='The Manager configuration file to use')

    args = parser.parse_args(WARN)

    manager = Manager(args.config_file)
    manager.validate_configs()


if __name__ == '__main__':
    main()



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import defaultdict
from requests import Session
import logging

from ..record import Record
from .base import BaseProvider


class DnsimpleClientException(Exception):
    pass


class DnsimpleClientNotFound(DnsimpleClientException):

    def __init__(self):
        super(DnsimpleClientNotFound, self).__init__('Not found')


class DnsimpleClientUnauthorized(DnsimpleClientException):

    def __init__(self):
        super(DnsimpleClientUnauthorized, self).__init__('Unauthorized')


class DnsimpleClient(object):
    BASE = 'https://api.dnsimple.com/v2/'

    def __init__(self, token, account):
        self.account = account
        sess = Session()
        sess.headers.update({'Authorization': 'Bearer {}'.format(token)})
        self._sess = sess

    def _request(self, method, path, params=None, data=None):
        url = '{}{}{}'.format(self.BASE, self.account, path)
        resp = self._sess.request(method, url, params=params, json=data)
        if resp.status_code == 401:
            raise DnsimpleClientUnauthorized()
        if resp.status_code == 404:
            raise DnsimpleClientNotFound()
        resp.raise_for_status()
        return resp

    def domain(self, name):
        path = '/domains/{}'.format(name)
        return self._request('GET', path).json()

    def domain_create(self, name):
        return self._request('POST', '/domains', data={'name': name})

    def records(self, zone_name):
        ret = []

        page = 1
        while True:
            data = self._request('GET', '/zones/{}/records'.format(zone_name),
                                 {'page': page}).json()
            ret += data['data']
            pagination = data['pagination']
            if page >= pagination['total_pages']:
                break
            page += 1

        return ret

    def record_create(self, zone_name, params):
        path = '/zones/{}/records'.format(zone_name)
        self._request('POST', path, data=params)

    def record_delete(self, zone_name, record_id):
        path = '/zones/{}/records/{}'.format(zone_name, record_id)
        self._request('DELETE', path)


class DnsimpleProvider(BaseProvider):
    '''
    Dnsimple provider using API v2

    dnsimple:
        class: octodns.provider.dnsimple.DnsimpleProvider
        # API v2 account access token (required)
        token: letmein
        # Your account number (required)
        account: 42
    '''
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS = set(('A', 'AAAA', 'ALIAS', 'CAA', 'CNAME', 'MX', 'NAPTR', 'NS',
                    'PTR', 'SPF', 'SRV', 'SSHFP', 'TXT'))

    def __init__(self, id, token, account, *args, **kwargs):
        self.log = logging.getLogger('DnsimpleProvider[{}]'.format(id))
        self.log.debug('__init__: id=%s, token=***, account=%s', id, account)
        super(DnsimpleProvider, self).__init__(id, *args, **kwargs)
        self._client = DnsimpleClient(token, account)

        self._zone_records = {}

    def _data_for_multiple(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': [r['content'] for r in records]
        }

    _data_for_A = _data_for_multiple
    _data_for_AAAA = _data_for_multiple
    _data_for_SPF = _data_for_multiple

    def _data_for_TXT(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            # escape semicolons
            'values': [r['content'].replace(';', '\\;') for r in records]
        }

    def _data_for_CAA(self, _type, records):
        values = []
        for record in records:
            flags, tag, value = record['content'].split(' ')
            values.append({
                'flags': flags,
                'tag': tag,
                'value': value[1:-1],
            })
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values
        }

    def _data_for_CNAME(self, _type, records):
        record = records[0]
        return {
            'ttl': record['ttl'],
            'type': _type,
            'value': '{}.'.format(record['content'])
        }

    _data_for_ALIAS = _data_for_CNAME

    def _data_for_MX(self, _type, records):
        values = []
        for record in records:
            values.append({
                'preference': record['priority'],
                'exchange': '{}.'.format(record['content'])
            })
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values
        }

    def _data_for_NAPTR(self, _type, records):
        values = []
        for record in records:
            try:
                order, preference, flags, service, regexp, replacement = \
                    record['content'].split(' ', 5)
            except ValueError:
                # their api will let you create invalid records, this
                # essentially handles that by ignoring them for values
                # purposes. That will cause updates to happen to delete them if
                # they shouldn't exist or update them if they're wrong
                continue
            values.append({
                'flags': flags[1:-1],
                'order': order,
                'preference': preference,
                'regexp': regexp[1:-1],
                'replacement': replacement,
                'service': service[1:-1],
            })
        return {
            'type': _type,
            'ttl': records[0]['ttl'],
            'values': values
        }

    def _data_for_NS(self, _type, records):
        values = []
        for record in records:
            content = record['content']
            if content[-1] != '.':
                content = '{}.'.format(content)
            values.append(content)
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values,
        }

    def _data_for_PTR(self, _type, records):
        record = records[0]
        return {
            'ttl': record['ttl'],
            'type': _type,
            'value': record['content']
        }

    def _data_for_SRV(self, _type, records):
        values = []
        for record in records:
            try:
                weight, port, target = record['content'].split(' ', 2)
            except ValueError:
                # see _data_for_NAPTR's continue
                continue
            values.append({
                'port': port,
                'priority': record['priority'],
                'target': '{}.'.format(target),
                'weight': weight
            })
        return {
            'type': _type,
            'ttl': records[0]['ttl'],
            'values': values
        }

    def _data_for_SSHFP(self, _type, records):
        values = []
        for record in records:
            try:
                algorithm, fingerprint_type, fingerprint = \
                    record['content'].split(' ', 2)
            except ValueError:
                # see _data_for_NAPTR's continue
                continue
            values.append({
                'algorithm': algorithm,
                'fingerprint': fingerprint,
                'fingerprint_type': fingerprint_type
            })
        return {
            'type': _type,
            'ttl': records[0]['ttl'],
            'values': values
        }

    def zone_records(self, zone):
        if zone.name not in self._zone_records:
            try:
                self._zone_records[zone.name] = \
                    self._client.records(zone.name[:-1])
            except DnsimpleClientNotFound:
                return []

        return self._zone_records[zone.name]

    def populate(self, zone, target=False, lenient=False):
        self.log.debug('populate: name=%s, target=%s, lenient=%s', zone.name,
                       target, lenient)

        values = defaultdict(lambda: defaultdict(list))
        for record in self.zone_records(zone):
            _type = record['type']
            if _type not in self.SUPPORTS:
                continue
            elif _type == 'TXT' and record['content'].startswith('ALIAS for'):
                # ALIAS has a "ride along" TXT record with 'ALIAS for XXXX',
                # we're ignoring it
                continue
            values[record['name']][record['type']].append(record)

        before = len(zone.records)
        for name, types in values.items():
            for _type, records in types.items():
                data_for = getattr(self, '_data_for_{}'.format(_type))
                record = Record.new(zone, name, data_for(_type, records),
                                    source=self, lenient=lenient)
                zone.add_record(record, lenient=lenient)

        exists = zone.name in self._zone_records
        self.log.info('populate:   found %s records, exists=%s',
                      len(zone.records) - before, exists)
        return exists

    def _params_for_multiple(self, record):
        for value in record.values:
            yield {
                'content': value,
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type,
            }

    _params_for_A = _params_for_multiple
    _params_for_AAAA = _params_for_multiple
    _params_for_NS = _params_for_multiple
    _params_for_SPF = _params_for_multiple

    def _params_for_TXT(self, record):
        for value in record.values:
            yield {
                # un-escape semicolons
                'content': value.replace('\\', ''),
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type,
            }

    def _params_for_CAA(self, record):
        for value in record.values:
            yield {
                'content': '{} {} "{}"'.format(value.flags, value.tag,
                                               value.value),
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_single(self, record):
        yield {
            'content': record.value,
            'name': record.name,
            'ttl': record.ttl,
            'type': record._type
        }

    _params_for_ALIAS = _params_for_single
    _params_for_CNAME = _params_for_single
    _params_for_PTR = _params_for_single

    def _params_for_MX(self, record):
        for value in record.values:
            yield {
                'content': value.exchange,
                'name': record.name,
                'priority': value.preference,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_NAPTR(self, record):
        for value in record.values:
            content = '{} {} "{}" "{}" "{}" {}' \
                .format(value.order, value.preference, value.flags,
                        value.service, value.regexp, value.replacement)
            yield {
                'content': content,
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_SRV(self, record):
        for value in record.values:
            yield {
                'content': '{} {} {}'.format(value.weight, value.port,
                                             value.target),
                'name': record.name,
                'priority': value.priority,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_SSHFP(self, record):
        for value in record.values:
            yield {
                'content': '{} {} {}'.format(value.algorithm,
                                             value.fingerprint_type,
                                             value.fingerprint),
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type
            }

    def _apply_Create(self, change):
        new = change.new
        params_for = getattr(self, '_params_for_{}'.format(new._type))
        for params in params_for(new):
            self._client.record_create(new.zone.name[:-1], params)

    def _apply_Update(self, change):
        self._apply_Delete(change)
        self._apply_Create(change)

    def _apply_Delete(self, change):
        existing = change.existing
        zone = existing.zone
        for record in self.zone_records(zone):
            if existing.name == record['name'] and \
               existing._type == record['type']:
                self._client.record_delete(zone.name[:-1], record['id'])

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug('_apply: zone=%s, len(changes)=%d', desired.name,
                       len(changes))

        domain_name = desired.name[:-1]
        try:
            self._client.domain(domain_name)
        except DnsimpleClientNotFound:
            self.log.debug('_apply:   no matching zone, creating domain')
            self._client.domain_create(domain_name)

        for change in changes:
            class_name = change.__class__.__name__
            getattr(self, '_apply_{}'.format(class_name))(change)

        # Clear out the cache if any
        self._zone_records.pop(desired.name, None)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
import re
from unittest import TestCase
from urlparse import urlparse

from requests import HTTPError
from requests_mock import ANY, mock as requests_mock

from octodns.provider.rackspace import RackspaceProvider
from octodns.record import Record
from octodns.zone import Zone

EMPTY_TEXT = '''
{
  "totalEntries" : 0,
  "records" : []
}
'''

with open('./tests/fixtures/rackspace-auth-response.json') as fh:
    AUTH_RESPONSE = fh.read()

with open('./tests/fixtures/rackspace-list-domains-response.json') as fh:
    LIST_DOMAINS_RESPONSE = fh.read()

with open('./tests/fixtures/rackspace-sample-recordset-page1.json') as fh:
    RECORDS_PAGE_1 = fh.read()

with open('./tests/fixtures/rackspace-sample-recordset-page2.json') as fh:
    RECORDS_PAGE_2 = fh.read()


class TestRackspaceProvider(TestCase):
    def setUp(self):
        self.maxDiff = 1000
        with requests_mock() as mock:
            mock.post(ANY, status_code=200, text=AUTH_RESPONSE)
            self.provider = RackspaceProvider('identity', 'test', 'api-key',
                                              '0')
            self.assertTrue(mock.called_once)

    def test_bad_auth(self):
        with requests_mock() as mock:
            mock.get(ANY, status_code=401, text='Unauthorized')

            with self.assertRaises(Exception) as ctx:
                zone = Zone('unit.tests.', [])
                self.provider.populate(zone)
            self.assertTrue('unauthorized' in ctx.exception.message)
            self.assertTrue(mock.called_once)

    def test_server_error(self):
        with requests_mock() as mock:
            mock.get(ANY, status_code=502, text='Things caught fire')

            with self.assertRaises(HTTPError) as ctx:
                zone = Zone('unit.tests.', [])
                self.provider.populate(zone)
            self.assertEquals(502, ctx.exception.response.status_code)
            self.assertTrue(mock.called_once)

    def test_nonexistent_zone(self):
        # Non-existent zone doesn't populate anything
        with requests_mock() as mock:
            mock.get(ANY, status_code=404,
                     json={'error': "Could not find domain 'unit.tests.'"})

            zone = Zone('unit.tests.', [])
            exists = self.provider.populate(zone)
            self.assertEquals(set(), zone.records)
            self.assertTrue(mock.called_once)
            self.assertFalse(exists)

    def test_multipage_populate(self):
        with requests_mock() as mock:
            mock.get(re.compile('domains$'), status_code=200,
                     text=LIST_DOMAINS_RESPONSE)
            mock.get(re.compile('records'), status_code=200,
                     text=RECORDS_PAGE_1)
            mock.get(re.compile('records.*offset=3'), status_code=200,
                     text=RECORDS_PAGE_2)

            zone = Zone('unit.tests.', [])
            self.provider.populate(zone)
            self.assertEquals(5, len(zone.records))

    def test_plan_disappearing_ns_records(self):
        expected = Zone('unit.tests.', [])
        expected.add_record(Record.new(expected, '', {
            'type': 'NS',
            'ttl': 600,
            'values': ['8.8.8.8.', '9.9.9.9.']
        }))
        expected.add_record(Record.new(expected, 'sub', {
            'type': 'NS',
            'ttl': 600,
            'values': ['8.8.8.8.', '9.9.9.9.']
        }))
        with requests_mock() as mock:
            mock.get(re.compile('domains$'), status_code=200,
                     text=LIST_DOMAINS_RESPONSE)
            mock.get(re.compile('records'), status_code=200, text=EMPTY_TEXT)

            plan = self.provider.plan(expected)
            self.assertTrue(mock.called)
            self.assertTrue(plan.exists)

            # OctoDNS does not propagate top-level NS records.
            self.assertEquals(1, len(plan.changes))

    def test_fqdn_a_record(self):
        expected = Zone('example.com.', [])
        # expected.add_record(Record.new(expected, 'foo', '1.2.3.4'))

        with requests_mock() as list_mock:
            list_mock.get(re.compile('domains$'), status_code=200,
                          text=LIST_DOMAINS_RESPONSE)
            list_mock.get(re.compile('records'), status_code=200,
                          json={'records': [
                              {'type': 'A',
                               'name': 'foo.example.com',
                               'id': 'A-111111',
                               'data': '1.2.3.4',
                               'ttl': 300}]})
            plan = self.provider.plan(expected)
            self.assertTrue(list_mock.called)
            self.assertEqual(1, len(plan.changes))
            self.assertTrue(
                plan.changes[0].existing.fqdn == 'foo.example.com.')

        with requests_mock() as mock:
            def _assert_deleting(request, context):
                parts = urlparse(request.url)
                self.assertEqual('id=A-111111', parts.query)

            mock.get(re.compile('domains$'), status_code=200,
                     text=LIST_DOMAINS_RESPONSE)
            mock.delete(re.compile('domains/.*/records?.*'), status_code=202,
                        text=_assert_deleting)
            self.provider.apply(plan)
            self.assertTrue(mock.called)

    def _test_apply_with_data(self, data):
        expected = Zone('unit.tests.', [])
        for record in data.OtherRecords:
            expected.add_record(
                Record.new(expected, record['subdomain'], record['data']))

        with requests_mock() as list_mock:
            list_mock.get(re.compile('domains$'), status_code=200,
                          text=LIST_DOMAINS_RESPONSE)
            list_mock.get(re.compile('records'), status_code=200,
                          json=data.OwnRecords)
            plan = self.provider.plan(expected)
            self.assertTrue(list_mock.called)
            if not data.ExpectChanges:
                self.assertFalse(plan)
                return

        with requests_mock() as mock:
            called = set()

            def make_assert_sending_right_body(expected):
                def _assert_sending_right_body(request, _context):
                    called.add(request.method)
                    if request.method != 'DELETE':
                        self.assertEqual(request.headers['content-type'],
                                         'application/json')
                        self.assertDictEqual(expected,
                                             json.loads(request.body))
                    else:
                        parts = urlparse(request.url)
                        self.assertEqual(expected, parts.query)
                    return ''

                return _assert_sending_right_body

            mock.get(re.compile('domains$'), status_code=200,
                     text=LIST_DOMAINS_RESPONSE)
            mock.post(re.compile('domains/.*/records$'), status_code=202,
                      text=make_assert_sending_right_body(
                          data.ExpectedAdditions))
            mock.delete(re.compile('domains/.*/records?.*'), status_code=202,
                        text=make_assert_sending_right_body(
                            data.ExpectedDeletions))
            mock.put(re.compile('domains/.*/records$'), status_code=202,
                     text=make_assert_sending_right_body(data.ExpectedUpdates))

            self.provider.apply(plan)
            self.assertTrue(data.ExpectedAdditions is None or "POST" in called)
            self.assertTrue(
                data.ExpectedDeletions is None or "DELETE" in called)
            self.assertTrue(data.ExpectedUpdates is None or "PUT" in called)

    def test_apply_no_change_empty(self):
        class TestData(object):
            OtherRecords = []
            OwnRecords = {
                "totalEntries": 0,
                "records": []
            }
            ExpectChanges = False
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_no_change_a_records(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'values': ['1.2.3.4', '1.2.3.5', '1.2.3.6']
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.5",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-333333",
                    "type": "A",
                    "data": "1.2.3.6",
                    "ttl": 300
                }]
            }
            ExpectChanges = False
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_no_change_a_records_cross_zone(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": 'foo',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'value': '1.2.3.4'
                    }
                },
                {
                    "subdomain": 'bar',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'value': '1.2.3.4'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "foo.unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "bar.unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }]
            }
            ExpectChanges = False
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_one_addition(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'value': '1.2.3.4'
                    }
                },
                {
                    "subdomain": 'foo',
                    "data": {
                        'type': 'NS',
                        'ttl': 300,
                        'value': 'ns.example.com.'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 0,
                "records": []
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "unit.tests",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "type": "NS",
                    "data": "ns.example.com",
                    "ttl": 300
                }]
            }
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_create_MX(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'MX',
                        'ttl': 300,
                        'value': {
                            'value': 'mail1.example.com.',
                            'priority': 1,
                        }
                    }
                },
                {
                    "subdomain": 'foo',
                    "data": {
                        'type': 'MX',
                        'ttl': 300,
                        'value': {
                            'value': 'mail2.example.com.',
                            'priority': 2
                        }
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 0,
                "records": []
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "foo.unit.tests",
                    "type": "MX",
                    "data": "mail2.example.com",
                    "priority": 2,
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "type": "MX",
                    "data": "mail1.example.com",
                    "priority": 1,
                    "ttl": 300
                }]
            }
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_additions_splatting(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'values': ['1.2.3.4', '1.2.3.5', '1.2.3.6']
                    }
                },
                {
                    "subdomain": 'foo',
                    "data": {
                        'type': 'NS',
                        'ttl': 300,
                        'values': ['ns1.example.com.', 'ns2.example.com.']
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 0,
                "records": []
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "unit.tests",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "type": "A",
                    "data": "1.2.3.5",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "type": "A",
                    "data": "1.2.3.6",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "type": "NS",
                    "data": "ns1.example.com",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "type": "NS",
                    "data": "ns2.example.com",
                    "ttl": 300
                }]
            }
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_additions_namespaced(self):
        class TestData(object):
            OtherRecords = [{
                "subdomain": 'foo',
                "data": {
                    'type': 'A',
                    'ttl': 300,
                    'value': '1.2.3.4'
                }
            }, {
                "subdomain": 'bar',
                "data": {
                    'type': 'A',
                    'ttl': 300,
                    'value': '1.2.3.4'
                }
            }, {
                "subdomain": 'foo',
                "data": {
                    'type': 'NS',
                    'ttl': 300,
                    'value': 'ns.example.com.'
                }
            }]
            OwnRecords = {
                "totalEntries": 0,
                "records": []
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "bar.unit.tests",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "type": "NS",
                    "data": "ns.example.com",
                    "ttl": 300
                }]
            }
            ExpectedDeletions = None
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_single_deletion(self):
        class TestData(object):
            OtherRecords = []
            OwnRecords = {
                "totalEntries": 1,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "id": "NS-111111",
                    "type": "NS",
                    "data": "ns.example.com",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = "id=A-111111&id=NS-111111"
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_deletions(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'value': '1.2.3.5'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.5",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-333333",
                    "type": "A",
                    "data": "1.2.3.6",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "id": "NS-111111",
                    "type": "NS",
                    "data": "ns.example.com",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = "id=A-111111&id=A-333333&id=NS-111111"
            ExpectedUpdates = {
                "records": [{
                    "name": "unit.tests",
                    "id": "A-222222",
                    "data": "1.2.3.5",
                    "ttl": 300
                }]
            }

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_deletions_cross_zone(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 300,
                        'value': '1.2.3.4'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "foo.unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.5",
                    "ttl": 300
                }, {
                    "name": "bar.unit.tests",
                    "id": "A-333333",
                    "type": "A",
                    "data": "1.2.3.6",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = "id=A-222222&id=A-333333"
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_delete_cname(self):
        class TestData(object):
            OtherRecords = []
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "foo.unit.tests",
                    "id": "CNAME-111111",
                    "type": "CNAME",
                    "data": "a.example.com",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = "id=CNAME-111111"
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_single_update(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 3600,
                        'value': '1.2.3.4'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 1,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = {
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "data": "1.2.3.4",
                    "ttl": 3600
                }]
            }

        return self._test_apply_with_data(TestData)

    def test_apply_update_TXT(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'TXT',
                        'ttl': 300,
                        'value': 'othervalue'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 1,
                "records": [{
                    "name": "unit.tests",
                    "id": "TXT-111111",
                    "type": "TXT",
                    "data": "somevalue",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "unit.tests",
                    "type": "TXT",
                    "data": "othervalue",
                    "ttl": 300
                }]
            }
            ExpectedDeletions = 'id=TXT-111111'
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_update_MX(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'MX',
                        'ttl': 300,
                        'value': {u'priority': 50, u'value': 'mx.test.com.'}
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 1,
                "records": [{
                    "name": "unit.tests",
                    "id": "MX-111111",
                    "type": "MX",
                    "priority": 20,
                    "data": "mx.test.com",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = {
                "records": [{
                    "name": "unit.tests",
                    "type": "MX",
                    "priority": 50,
                    "data": "mx.test.com",
                    "ttl": 300
                }]
            }
            ExpectedDeletions = 'id=MX-111111'
            ExpectedUpdates = None

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_updates(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": '',
                    "data": {
                        'type': 'A',
                        'ttl': 3600,
                        'values': ['1.2.3.4', '1.2.3.5', '1.2.3.6']
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 3,
                "records": [{
                    "name": "unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.5",
                    "ttl": 300
                }, {
                    "name": "unit.tests",
                    "id": "A-333333",
                    "type": "A",
                    "data": "1.2.3.6",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = {
                "records": [{
                    "name": "unit.tests",
                    "id": "A-222222",
                    "data": "1.2.3.5",
                    "ttl": 3600
                }, {
                    "name": "unit.tests",
                    "id": "A-111111",
                    "data": "1.2.3.4",
                    "ttl": 3600
                }, {
                    "name": "unit.tests",
                    "id": "A-333333",
                    "data": "1.2.3.6",
                    "ttl": 3600
                }]
            }

        return self._test_apply_with_data(TestData)

    def test_apply_multiple_updates_cross_zone(self):
        class TestData(object):
            OtherRecords = [
                {
                    "subdomain": 'foo',
                    "data": {
                        'type': 'A',
                        'ttl': 3600,
                        'value': '1.2.3.4'
                    }
                },
                {
                    "subdomain": 'bar',
                    "data": {
                        'type': 'A',
                        'ttl': 3600,
                        'value': '1.2.3.4'
                    }
                }
            ]
            OwnRecords = {
                "totalEntries": 2,
                "records": [{
                    "name": "foo.unit.tests",
                    "id": "A-111111",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }, {
                    "name": "bar.unit.tests",
                    "id": "A-222222",
                    "type": "A",
                    "data": "1.2.3.4",
                    "ttl": 300
                }]
            }
            ExpectChanges = True
            ExpectedAdditions = None
            ExpectedDeletions = None
            ExpectedUpdates = {
                "records": [{
                    "name": "bar.unit.tests",
                    "id": "A-222222",
                    "data": "1.2.3.4",
                    "ttl": 3600
                }, {
                    "name": "foo.unit.tests",
                    "id": "A-111111",
                    "data": "1.2.3.4",
                    "ttl": 3600
                }]
            }

        return self._test_apply_with_data(TestData)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals


class BaseSource(object):

    def __init__(self, id):
        self.id = id
        if not getattr(self, 'log', False):
            raise NotImplementedError('Abstract base class, log property '
                                      'missing')
        if not hasattr(self, 'SUPPORTS_GEO'):
            raise NotImplementedError('Abstract base class, SUPPORTS_GEO '
                                      'property missing')
        if not hasattr(self, 'SUPPORTS'):
            raise NotImplementedError('Abstract base class, SUPPORTS '
                                      'property missing')

    @property
    def SUPPORTS_DYNAMIC(self):
        return False

    def populate(self, zone, target=False, lenient=False):
        '''
        Loads all records the provider knows about for the provided zone

        When `target` is True the populate call is being made to load the
        current state of the provider.

        When `lenient` is True the populate call may skip record validation and
        do a "best effort" load of data. That will allow through some common,
        but not best practices stuff that we otherwise would reject. E.g. no
        trailing . or mising escapes for ;.

        When target is True (loading current state) this method should return
        True if the zone exists or False if it does not.
        '''
        raise NotImplementedError('Abstract base class, populate method '
                                  'missing')

    def supports(self, record):
        return record._type in self.SUPPORTS

    def __repr__(self):
        return self.__class__.__name__
